# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from collections.abc import Sequence as Seq
from collections.abc import Sized
import logging
import math
from typing import (
    Any, Collection, Dict, List, Optional, Sequence, Tuple, Union,
    cast, TYPE_CHECKING)
from typing_extensions import TypeGuard
import numpy
from numpy.typing import NDArray
import scipy.stats
from pyNN.space import Grid2D, Grid3D, BaseStructure
from spinn_utilities.log import FormatAdapter
from spinn_utilities.overrides import overrides
from spinn_utilities.ranged import RangeDictionary, RangedList
from spinn_utilities.config_holder import get_config_int
from pacman.model.graphs.application import ApplicationEdge
from pacman.model.graphs.common import Slice
from pacman.model.resources import AbstractSDRAM, ConstantSDRAM
from pacman.model.partitioner_interfaces import LegacyPartitionerAPI
from pacman.model.partitioner_splitters import AbstractSplitterCommon
from spinn_front_end_common.interface.buffer_management import (
    recording_utilities)
from spinn_front_end_common.utilities.constants import (
    SYSTEM_BYTES_REQUIREMENT)
from spinn_front_end_common.interface.profiling import profile_utils
from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.models.common import MultiSpikeRecorder
from spynnaker.pyNN.utilities.utility_calls import create_mars_kiss_seeds
from spynnaker.pyNN.models.abstract_models import SupportsStructure
from spynnaker.pyNN.models.common import (
    ParameterHolder, PopulationApplicationVertex)
from spynnaker.pyNN.models.common.types import Names
from spynnaker.pyNN.utilities.buffer_data_type import BufferDataType
from spynnaker.pyNN.models.neuron.synapse_dynamics.types import (
    ConnectionsArray)
from .spike_source_poisson_machine_vertex import (
    SpikeSourcePoissonMachineVertex, _flatten, get_rates_bytes,
    get_sdram_edge_params_bytes, get_expander_rates_bytes, get_params_bytes)
if TYPE_CHECKING:
    from spinn_utilities.ranged.abstract_sized import Selector
    from .spike_source_poisson import SpikeSourcePoisson
    from .spike_source_poisson_variable import SpikeSourcePoissonVariable
    from spynnaker.pyNN.models.neural_projections import SynapseInformation
    from spynnaker.pyNN.models.projection import Projection
    from spynnaker.pyNN.models.common.types import Values

logger = FormatAdapter(logging.getLogger(__name__))

# uint32_t n_rates; uint32_t index
PARAMS_WORDS_PER_NEURON = 2

# start_scaled, end_scaled, next_scaled, is_fast_source, exp_minus_lambda,
# sqrt_lambda, isi_val, time_to_spike
PARAMS_WORDS_PER_RATE = 8

SLOW_RATE_PER_TICK_CUTOFF = (
    SpikeSourcePoissonMachineVertex.SLOW_RATE_PER_TICK_CUTOFF)

OVERFLOW_TIMESTEPS_FOR_SDRAM = 5

# The microseconds per timestep will be divided by this to get the max offset
_MAX_OFFSET_DENOMINATOR = 10

# Indicates a duration that never ends
DURATION_FOREVER = 0xFFFFFFFF


def _is_list_of_lists(value: Any) -> TypeGuard[
        Sequence[Sequence[Union[int, float]]]]:
    return isinstance(value, (Seq, numpy.ndarray)) and isinstance(
        value[0], (Seq, numpy.ndarray))


def _normalize_rates(
        rate: Union[float, Sequence[float], None],
        rates: Union[Sequence[float], NDArray[numpy.floating], None]
        ) -> Union[NDArray[numpy.floating], List[NDArray[numpy.floating]]]:
    if rates is None:
        if isinstance(rate, (Sequence, numpy.ndarray)):
            # Single rate per neuron for whole simulation
            return [numpy.array([r]) for r in rate]
        else:
            # Single rate for all neurons for whole simulation
            return numpy.array([rate])
    elif _is_list_of_lists(rates):
        # Convert each list to numpy array
        return [numpy.array(r) for r in rates]
    else:
        return numpy.array(rates)


def _normalize_times(
        time: Union[int, Sequence[int], None],
        times: Union[Sequence[int], NDArray[numpy.integer], None]
        ) -> Union[NDArray[numpy.integer], List[NDArray[numpy.integer]], None]:
    if times is None:
        if time is None:
            return None
        if isinstance(time, Sequence):
            # Single time per neuron for whole simulation
            return [numpy.array([r]) for r in time]
        else:
            # Single time for all neurons for whole simulation
            return numpy.array([time])
    elif _is_list_of_lists(times):
        # Convert each list to numpy array
        return [numpy.array(r) for r in times]
    else:
        return numpy.array(times)


def is_iterable(value: Values) -> TypeGuard[
        Union[Sequence[float], NDArray[numpy.floating]]]:
    """
    Check that the Value is iterable.

    :param value: The value to check.
    :returns: True if Value is iterable, False otherwise.
    """
    return hasattr(value, "__iter__")


class SpikeSourcePoissonVertex(
        PopulationApplicationVertex,
        LegacyPartitionerAPI, SupportsStructure):
    """
    A SpiNNaker vertex that is a Poisson-distributed Spike source.
    """

    __slots__ = (
        "__last_rate_read_time",
        "__model",
        "__model_name",
        "__n_atoms",
        "__rng",
        "__seed",
        "__spike_recorder",
        "__kiss_seed",  # dict indexed by vertex slice
        "__max_rate",
        "__max_n_rates",
        "__n_profile_samples",
        "__data",
        "__is_variable_rate",
        "__outgoing_projections",
        "__incoming_control_edge",
        "__structure",
        "__allowed_parameters",
        "__n_colour_bits")

    SPIKE_RECORDING_REGION_ID = 0

    def __init__(
            self, n_neurons: int, label: str, seed: Optional[int],
            max_atoms_per_core: Optional[Union[int, Tuple[int, ...]]],
            model: Union[SpikeSourcePoisson, SpikeSourcePoissonVariable],
            rate: Union[float, Sequence[float], None] = None,
            start: Union[int, Sequence[int], None] = None,
            duration: Union[int, Sequence[int], None] = None,
            rates: Union[
                Sequence[float], NDArray[numpy.floating], None] = None,
            starts: Union[Sequence[int], NDArray[numpy.integer], None] = None,
            durations: Union[
                Sequence[int], NDArray[numpy.integer], None] = None,
            max_rate: Optional[float] = None,
            splitter: Optional[AbstractSplitterCommon] = None,
            n_colour_bits: Optional[int] = None):
        """
        :param n_neurons: The number of neurons in this vertex.
        :param label: The optional name of the vertex.
        :param max_atoms_per_core: The max number of atoms that can be
            placed on a core for each dimension, used in partitioning.
            If the vertex is n-dimensional, with n > 1, the value must be a
            tuple with a value for each dimension.  If it is single-dimensional
            the value can be a 1-tuple or an int.
        :param model: The model to get the parameters from
        :param rate: The spike rate of all neuron
        :param start: The start time of spikes on all neurons
        :param duration: The duration of spikes on all neurons
        :param rates: The spike rate of each neuron
        :param starts: The start time of spikes on each neuron
        :param durations: The duration of spikes on each neuron
        :param max_rate: The maximum number of spikes
            for any neuron at any timestamp
        :param splitter: The splitter object needed for this vertex.
            Leave as `None` to delegate the choice of splitter to the selector.
        :param n_colour_bits:
        """
        super().__init__(label, max_atoms_per_core, splitter)

        # atoms params
        self.__n_atoms = self.round_n_atoms(n_neurons, "n_neurons")
        self.__model_name = "SpikeSourcePoisson"
        self.__model = model
        self.__seed = seed
        self.__kiss_seed: Dict[Slice, Tuple[int, ...]] = dict()

        self.__spike_recorder = MultiSpikeRecorder()

        # Check for disallowed pairs of parameters
        if (rates is not None) and (rate is not None):
            raise ValueError("Exactly one of rate and rates can be specified")
        if (starts is not None) and (start is not None):
            raise ValueError(
                "Exactly one of start and starts can be specified")
        if (durations is not None) and (duration is not None):
            raise ValueError(
                "Exactly one of duration and durations can be specified")
        if rate is None and rates is None:
            raise ValueError("One of rate or rates must be specified")

        # Normalise the parameters
        self.__is_variable_rate = rates is not None
        _rates = _normalize_rates(rate, rates)
        _starts = _normalize_times(start, starts)
        _durations = _normalize_times(duration, durations)
        if _durations is None:
            if _is_list_of_lists(_rates):
                _durations = [numpy.array([DURATION_FOREVER for _r in _rate])
                              for _rate in _rates]
            else:
                _durations = numpy.array(
                    [DURATION_FOREVER for _rate in _rates])

        # Check that there is either one list for all neurons,
        # or one per neuron
        if _is_list_of_lists(_rates) and len(_rates) != n_neurons:
            raise ValueError(
                "Must specify one rate for all neurons or one per neuron")
        if _is_list_of_lists(_starts) and len(_starts) != n_neurons:
            raise ValueError(
                "Must specify one start for all neurons or one per neuron")
        if _is_list_of_lists(_durations) and len(_durations) != n_neurons:
            raise ValueError(
                "Must specify one duration for all neurons or one per neuron")

        # Check that for each rate there is a start and duration if needed
        # TODO: Could be more efficient for case where parameters are not one
        #       per neuron
        for i in range(n_neurons):
            rate_set = _rates[i] if _is_list_of_lists(_rates) else _rates
            if not isinstance(rate_set, Sized):
                raise ValueError("Multiple rates must be a list")
            if starts is None and len(rate_set) > 1:
                raise ValueError(
                    "When multiple rates are specified,"
                    " each must have a start")
            elif _starts is not None:
                start_set = (
                    _starts[i] if _is_list_of_lists(_starts) else _starts)
                if len(start_set) != len(rate_set):
                    raise ValueError("Each rate must have a start")
                if any(s is None for s in start_set):
                    raise ValueError("Start must not be None")
            if _durations is not None:
                duration_set = (
                    _durations[i] if _is_list_of_lists(_durations)
                    else _durations)
                if len(duration_set) != len(rate_set):
                    raise ValueError("Each rate must have its own duration")

        self.__data: RangeDictionary[
            Union[NDArray[numpy.floating], NDArray[numpy.integer]]
            ] = RangeDictionary(n_neurons)
        rates_list: RangedList = RangedList(
            n_neurons, _rates,
            use_list_as_value=not _is_list_of_lists(_rates))
        self.__data["rates"] = rates_list
        self.__data["starts"] = RangedList(
            n_neurons, _starts,
            use_list_as_value=not _is_list_of_lists(_starts))
        self.__data["durations"] = RangedList(
            n_neurons, _durations,
            use_list_as_value=not _is_list_of_lists(_durations))
        self.__rng = numpy.random.RandomState(seed)

        self.__n_profile_samples = get_config_int(
            "Reports", "n_profile_samples")

        # Prepare for recording, and to get spikes
        self.__spike_recorder = MultiSpikeRecorder()

        if max_rate is None:
            all_rates: List[numpy.floating] = list(
                _flatten(self.__data["rates"]))
            self.__max_rate = numpy.amax(all_rates) if all_rates else 0
        else:
            self.__max_rate = max_rate
        self.__max_n_rates = max(len(r) for r in rates_list)

        # Keep track of how many outgoing projections exist
        self.__outgoing_projections: List[Projection] = list()
        self.__incoming_control_edge: Optional[ApplicationEdge] = None

        self.__structure: Optional[BaseStructure] = None

        if self.__is_variable_rate:
            self.__allowed_parameters = frozenset(
                {"rates", "durations", "starts"})
        else:
            self.__allowed_parameters = frozenset(
                {"rate", "duration", "start"})

        self.__last_rate_read_time: Optional[float] = None

        if n_colour_bits is None:
            n_colour_bits = get_config_int("Simulation", "n_colour_bits")
        self.__n_colour_bits = n_colour_bits

    @overrides(SupportsStructure.set_structure)
    def set_structure(self, structure: BaseStructure) -> None:
        self.__structure = structure

    @property
    def rates(self) -> RangedList[NDArray[numpy.floating]]:
        """
        Get the rates.
        """
        # UGH! Mypy has been defeated!
        return cast(Any, self.__data["rates"])

    def add_outgoing_projection(self, projection: Projection) -> None:
        """
        Add an outgoing projection from this vertex.

        :param projection: The projection to add
        """
        self.__outgoing_projections.append(projection)

    @property
    def outgoing_projections(self) -> Sequence[Projection]:
        """
        The projections outgoing from this vertex.
        """
        return self.__outgoing_projections

    @property
    def n_profile_samples(self) -> int:
        """
        The n_profile_samples read from the config
        """
        return self.__n_profile_samples

    @property
    def time_to_spike(self) -> RangedList:
        """
        The "time_to_spike range list.
        """
        return self.__data["time_to_spike"]

    def __read_parameters_now(self) -> None:
        # If we already read the parameters at this time, don't do it again
        current_time = SpynnakerDataView().get_current_run_time_ms()
        if self.__last_rate_read_time == current_time:
            return

        self.__last_rate_read_time = current_time
        for m_vertex in self.machine_vertices:
            placement = SpynnakerDataView.get_placement_of_vertex(m_vertex)
            m_vertex.read_parameters_from_machine(placement)

    def __read_parameter(self, name: str, selector: Selector) -> Sequence:
        if (SpynnakerDataView.is_ran_last() and
                SpynnakerDataView.has_transceiver()):
            self.__read_parameters_now()
        return self.__data[self.__full_name(name)].get_values(selector)

    def __full_name(self, name: str) -> str:
        if self.__is_variable_rate:
            return name
        return f"{name}s"

    @overrides(PopulationApplicationVertex.get_parameter_values)
    def get_parameter_values(
            self, names: Names, selector: Selector = None) -> ParameterHolder:
        self._check_parameters(names, self.__allowed_parameters)
        return ParameterHolder(names, self.__read_parameter, selector)

    @overrides(PopulationApplicationVertex.set_parameter_values)
    def set_parameter_values(
            self, name: str, value: Values, selector: Selector = None) -> None:
        self._check_parameters(name, self.__allowed_parameters)
        if self.__is_variable_rate:
            raise KeyError(f"Cannot set the {name} of a variable rate Poisson")

        # If we have just run, we need to read parameters to avoid overwrite
        if SpynnakerDataView().is_ran_last():
            self.__read_parameters_now()
        for m_vertex in self.machine_vertices:
            m_vertex.set_rate_changed()

        # Must be parameter without the s
        fixed_name = f"{name}s"
        if is_iterable(value):
            # Single start per neuron for whole simulation
            self.__data[fixed_name].set_value_by_selector(
                selector, [numpy.array([s]) for s in value])
        else:
            # Single start for all neurons for whole simulation
            self.__data[fixed_name].set_value_by_selector(
                selector, numpy.array([value]), use_list_as_value=True)

    @overrides(PopulationApplicationVertex.get_parameters)
    def get_parameters(self) -> List[str]:
        return list(self.__allowed_parameters)

    @overrides(PopulationApplicationVertex.get_units)
    def get_units(self, name: str) -> str:
        if name == "spikes":
            return ""
        if name == "rates" or name == "rates":
            return "Hz"
        if (name == "duration" or name == "start" or name == "durations" or
                name == "starts"):
            return "ms"
        raise KeyError(f"Units for {name} unknown")

    @overrides(PopulationApplicationVertex.get_recordable_variables)
    def get_recordable_variables(self) -> List[str]:
        return ["spikes"]

    def get_buffer_data_type(self, name: str) -> BufferDataType:
        if name == "spikes":
            return BufferDataType.MULTI_SPIKES
        raise KeyError(f"Cannot record {name}")

    def get_recording_region(self, name: str) -> int:
        if name != "spikes":
            raise KeyError(f"Cannot record {name}")
        return 0

    @overrides(PopulationApplicationVertex.set_recording)
    def set_recording(
            self, name: str, sampling_interval: Optional[float] = None,
            indices: Optional[Collection[int]] = None) -> None:
        if name != "spikes":
            raise KeyError(f"Cannot record {name}")
        if sampling_interval is not None:
            logger.warning("Sampling interval currently not supported for "
                           "SpikeSourcePoisson so being ignored")
        if indices is not None:
            logger.warning("Indices currently not supported for "
                           "SpikeSourcePoisson so being ignored")
        if not self.__spike_recorder.record:
            SpynnakerDataView.set_requires_mapping()
        self.__spike_recorder.record = True

    @overrides(PopulationApplicationVertex.get_recording_variables)
    def get_recording_variables(self) -> List[str]:
        if self.__spike_recorder.record:
            return ["spikes"]
        return []

    @overrides(PopulationApplicationVertex.set_not_recording)
    def set_not_recording(self, name: str,
                          indices: Optional[Collection[int]] = None) -> None:
        if name != "spikes":
            raise KeyError(f"Cannot record {name}")
        if indices is not None:
            logger.warning("Indices currently not supported for "
                           "SpikeSourceArray so being ignored")
        self.__spike_recorder.record = False

    @overrides(PopulationApplicationVertex.get_sampling_interval_ms)
    def get_sampling_interval_ms(self, name: str) -> int:
        # TODO microseconds or milliseconds?
        if name != "spikes":
            raise KeyError(f"Cannot record {name}")
        return SpynnakerDataView.get_simulation_time_step_us()

    @overrides(PopulationApplicationVertex.get_data_type)
    def get_data_type(self, name: str) -> None:
        if name != "spikes":
            raise KeyError(f"Cannot record {name}")
        return None

    @overrides(PopulationApplicationVertex.get_neurons_recording)
    def get_neurons_recording(self, name: str, vertex_slice: Slice) -> NDArray:
        if name != "spikes":
            raise KeyError(f"Cannot record {name}")
        return vertex_slice.get_raster_ids()

    def max_spikes_per_ts(self) -> float:
        """
        Compute the maximum spike rate.

        :return: The maximum number of spikes per simulation timestep.
        """
        ts_per_second = SpynnakerDataView.get_simulation_time_step_per_s()
        if float(self.__max_rate) / ts_per_second < SLOW_RATE_PER_TICK_CUTOFF:
            return 1.0

        # Experiments show at 1000 this result is typically higher than actual
        chance_ts = 1000
        max_spikes_per_ts = scipy.stats.poisson.ppf(
            1.0 - (1.0 / float(chance_ts)),
            float(self.__max_rate) / ts_per_second)
        return int(math.ceil(max_spikes_per_ts)) + 1.0

    def get_recording_sdram_usage(self, vertex_slice: Slice) -> AbstractSDRAM:
        """
        :param vertex_slice: Slice to get cost for
        :returns: SDRAm cost for recording
        """
        variable_sdram = self.__spike_recorder.get_sdram_usage_in_bytes(
            vertex_slice.n_atoms, self.max_spikes_per_ts())
        constant_sdram = ConstantSDRAM(math.ceil(
            variable_sdram.per_timestep * OVERFLOW_TIMESTEPS_FOR_SDRAM))
        return variable_sdram + constant_sdram

    @overrides(LegacyPartitionerAPI.get_sdram_used_by_atoms)
    def get_sdram_used_by_atoms(self, vertex_slice: Slice) -> AbstractSDRAM:
        poisson_params_sz = get_params_bytes(vertex_slice.n_atoms)
        poisson_rates_sz = get_rates_bytes(
            vertex_slice.n_atoms, vertex_slice.n_atoms * self.__max_n_rates)
        poisson_expander_sz = get_expander_rates_bytes(
            vertex_slice.n_atoms, vertex_slice.n_atoms * self.__max_n_rates)
        sdram_sz = get_sdram_edge_params_bytes(vertex_slice)
        other = ConstantSDRAM(
            SYSTEM_BYTES_REQUIREMENT +
            SpikeSourcePoissonMachineVertex.get_provenance_data_size(1) +
            poisson_params_sz + poisson_rates_sz + poisson_expander_sz +
            recording_utilities.get_recording_header_size(1) +
            recording_utilities.get_recording_data_constant_size(1) +
            profile_utils.get_profile_region_size(self.__n_profile_samples) +
            sdram_sz)

        recording = self.get_recording_sdram_usage(vertex_slice)
        return recording + other

    @property
    @overrides(PopulationApplicationVertex.n_atoms)
    def n_atoms(self) -> int:
        return self.__n_atoms

    @property
    @overrides(PopulationApplicationVertex.atoms_shape)
    def atoms_shape(self) -> Tuple[int, ...]:
        if isinstance(self.__structure, (Grid2D, Grid3D)):
            return self.__structure.calculate_size(self.__n_atoms)
        return super().atoms_shape

    @overrides(LegacyPartitionerAPI.create_machine_vertex)
    def create_machine_vertex(
            self, vertex_slice: Slice, sdram: AbstractSDRAM,
            label: Optional[str] = None) -> SpikeSourcePoissonMachineVertex:
        return SpikeSourcePoissonMachineVertex(
            sdram, self.__spike_recorder.record,
            label, self, vertex_slice)

    @property
    def max_rate(self) -> float:
        """
        The highest rate or 0 if no rate set.
        """
        return float(self.__max_rate)

    @property
    def max_n_rates(self) -> int:
        """
        the long length of any rates list.
        """
        return self.__max_n_rates

    @property
    def seed(self) -> Optional[int]:
        """
        The seed set if any.
        """
        return self.__seed

    @seed.setter
    def seed(self, seed: int) -> None:
        self.__seed = seed
        self.__kiss_seed = dict()
        self.__rng = numpy.random.RandomState(seed)

    def kiss_seed(self, vertex_slice: Slice) -> Tuple[int, ...]:
        """
        The seed for this vertex slice.

        Generates and checks that the seed values generated by the given
        random number generator or seed to a random number generator are
        suitable for use as a mars 64 kiss seed.

        :param vertex_slice:
        :return: a list of 4 integers which are used by the mars64 kiss random
            number generator for seeds.
        """
        if vertex_slice not in self.__kiss_seed:
            self.__kiss_seed[vertex_slice] = create_mars_kiss_seeds(self.__rng)
        return self.__kiss_seed[vertex_slice]

    def update_kiss_seed(
            self, vertex_slice: Slice, seed: Sequence[int]) -> None:
        """
        Updates a KISS seed from the machine.

        :param vertex_slice: the vertex slice to update seed of
        :param seed: the seed
        """
        self.__kiss_seed[vertex_slice] = tuple(seed)

    def clear_spike_recording(self) -> None:
        """
        Clears the spike data from the buffer manager for this vertex.
        """
        buffer_manager = SpynnakerDataView.get_buffer_manager()
        for machine_vertex in self.machine_vertices:
            placement = SpynnakerDataView.get_placement_of_vertex(
                machine_vertex)
            buffer_manager.clear_recorded_data(
                placement.x, placement.y, placement.p,
                SpikeSourcePoissonVertex.SPIKE_RECORDING_REGION_ID)

    def describe(
            self) -> Dict[str, Union[str, ParameterHolder, Dict[str, Any]]]:
        """
        Return a human-readable description of the cell or synapse type.

        The output may be customised by specifying a different template
        together with an associated template engine
        (see :py:mod:`pyNN.descriptions`).

        If template is `None`, then a dictionary containing the template
        context will be returned.

        :returns:  human-readable description of the vertex
        """
        parameters = self.get_parameter_values(self.__model.default_parameters)

        return {
            "name": self.__model_name,
            "default_parameters": self.__model.default_parameters,
            "default_initial_values": self.__model.default_parameters,
            "parameters": parameters,
        }

    def set_live_poisson_control_edge(self, edge: ApplicationEdge) -> None:
        """
        Sets the poisson generator.

        :param edge:
        :raises ValueError: if already set
        """
        if self.__incoming_control_edge is not None:
            raise ValueError(
                "The Poisson generator can only be controlled by one source")
        self.__incoming_control_edge = edge

    @property
    def incoming_control_edge(self) -> Optional[ApplicationEdge]:
        """
        The live poisson control edge/ generator is set
        """
        return self.__incoming_control_edge

    @property
    def data(self) -> RangeDictionary[
            Union[NDArray[numpy.floating], NDArray[numpy.integer]]]:
        """
        A dictionary holding all the data as ranges
        """
        return self.__data

    @property
    def n_colour_bits(self) -> int:
        return self.__n_colour_bits

    def read_connections(
            self, synapse_info: SynapseInformation) -> List[ConnectionsArray]:
        """ Read Poisson connections from the machine

        :param synapse_info: The synapse information of the data being read
        :return: The set of connections from all machine vertices
        """
        connections = list()
        for m_vertex in self.machine_vertices:
            connections.append(m_vertex.read_connections(synapse_info))
        return connections



# Copyright (c) 2020 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Iterable, Sequence, Tuple

from spinn_utilities.overrides import overrides

from pacman.exceptions import (
    PacmanConfigurationException, PacmanInvalidParameterException)
from pacman.model.graphs.application import ApplicationVertex
from pacman.model.graphs.machine import MachineVertex
from pacman.model.graphs.common import Slice
from pacman.model.partitioner_splitters import AbstractSplitterCommon
from pacman.model.resources import AbstractSDRAM, ConstantSDRAM
from pacman.utilities.utility_objs.chip_counter import ChipCounter

from spinn_front_end_common.utilities.constants import (
    SYSTEM_BYTES_REQUIREMENT, BYTES_PER_WORD)

from spynnaker.pyNN.models.utility_models.delays import (
    DelayExtensionVertex, DelayExtensionMachineVertex)


class SplitterDelayVertexSlice(AbstractSplitterCommon[DelayExtensionVertex]):
    """
    Handles the splitting of the :py:class:`DelayExtensionVertex`
    via slice logic.
    """

    __slots__ = ("_machine_vertex_by_slice", )

    _EXPANDER_BASE_PARAMS_SIZE = 3 * BYTES_PER_WORD

    NEED_EXACT_ERROR_MESSAGE = (
        "DelayExtensionsSplitters need exact incoming slices. Please fix "
        "and try again")

    DELAY_RECORDING_ERROR = (
        "The delay extensions does not record any variables. Therefore "
        "asking for them is deemed an error.")

    def __init__(self) -> None:
        super().__init__()
        self._machine_vertex_by_slice: Dict[
            Slice, DelayExtensionMachineVertex] = dict()

    @overrides(AbstractSplitterCommon.get_out_going_vertices)
    def get_out_going_vertices(
            self, partition_id: str) -> Sequence[MachineVertex]:
        return list(self.governed_app_vertex.machine_vertices)

    @overrides(AbstractSplitterCommon.get_in_coming_vertices)
    def get_in_coming_vertices(
            self, partition_id: str) -> Sequence[MachineVertex]:
        return list(self.governed_app_vertex.machine_vertices)

    @overrides(AbstractSplitterCommon.get_source_specific_in_coming_vertices)
    def get_source_specific_in_coming_vertices(
            self, source_vertex: ApplicationVertex,
            partition_id: str) -> Sequence[Tuple[
                DelayExtensionMachineVertex, Sequence[MachineVertex]]]:
        # Only connect to the source that matches the slice
        return [
            (self._machine_vertex_by_slice[m_vertex.vertex_slice], [m_vertex])
            for m_vertex in source_vertex.splitter.get_out_going_vertices(
                partition_id)]

    @overrides(AbstractSplitterCommon.create_machine_vertices)
    def create_machine_vertices(self, chip_counter: ChipCounter) -> None:
        source_app_vertex = self.governed_app_vertex.source_vertex
        slices = source_app_vertex.splitter.get_out_going_slices()

        # create vertices correctly
        for vertex_slice in slices:
            vertex = self.create_machine_vertex(
                source_app_vertex, vertex_slice)
            self.governed_app_vertex.remember_machine_vertex(vertex)
            chip_counter.add_core(vertex.sdram_required)

    @overrides(AbstractSplitterCommon.get_in_coming_slices)
    def get_in_coming_slices(self) -> Sequence[Slice]:
        other_splitter = self.governed_app_vertex.source_vertex.splitter
        return other_splitter.get_in_coming_slices()

    @overrides(AbstractSplitterCommon.get_out_going_slices)
    def get_out_going_slices(self) -> Sequence[Slice]:
        other_splitter = self.governed_app_vertex.source_vertex.splitter
        return other_splitter.get_out_going_slices()

    @overrides(AbstractSplitterCommon.set_governed_app_vertex)
    def set_governed_app_vertex(
            self, app_vertex: DelayExtensionVertex) -> None:
        super().set_governed_app_vertex(app_vertex)
        if not isinstance(app_vertex, DelayExtensionVertex):
            raise PacmanConfigurationException(
                f"The vertex {app_vertex} cannot be supported by the "
                "SplitterDelayVertexSlice as the only vertex supported by "
                "this splitter is a DelayExtensionVertex. Please use the "
                "correct splitter for your vertex and try again.")

    def create_machine_vertex(
            self, source_app_vertex: ApplicationVertex,
            vertex_slice: Slice) -> DelayExtensionMachineVertex:
        """
        Creates a delay extension machine vertex and adds to the tracker.

        :param source_app_vertex: The source of the delay
        :param vertex_slice: The neurons this Machine Vertex should cover.
        :return: machine vertex
        """
        label = f"Delay extension for {source_app_vertex}"
        sdram = self.get_sdram_used_by_atoms()

        machine_vertex = DelayExtensionMachineVertex(
            sdram, label, vertex_slice, self.governed_app_vertex)

        self._machine_vertex_by_slice[vertex_slice] = machine_vertex
        return machine_vertex

    def get_sdram_used_by_atoms(self) -> AbstractSDRAM:
        """
        :returns: The amount of SDRAM used by the delay extension.
        """
        return ConstantSDRAM(
            SYSTEM_BYTES_REQUIREMENT +
            self.governed_app_vertex.delay_params_size() +
            DelayExtensionMachineVertex.get_provenance_data_size(
                DelayExtensionMachineVertex.N_EXTRA_PROVENANCE_DATA_ENTRIES))

    @overrides(AbstractSplitterCommon.machine_vertices_for_recording)
    def machine_vertices_for_recording(
            self, variable_to_record: str) -> Iterable[MachineVertex]:
        raise PacmanInvalidParameterException(
            variable_to_record, variable_to_record, self.DELAY_RECORDING_ERROR)

    @overrides(AbstractSplitterCommon.reset_called)
    def reset_called(self) -> None:
        self._machine_vertex_by_slice = dict()

    def get_machine_vertex(
            self, vertex_slice: Slice) -> DelayExtensionMachineVertex:
        """
        :param vertex_slice: The slice to get the data for
        :returns: A delay extension machine vertex for the vertex slice.
        """
        return self._machine_vertex_by_slice[vertex_slice]



# Copyright (c) 2020 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import defaultdict
import logging
import math
from typing import Dict, List, Optional, Sequence, Set, Tuple, cast

from numpy import floating
from numpy.typing import NDArray

from spinn_utilities.overrides import overrides
from spinn_utilities.log import FormatAdapter
from spinn_utilities.ordered_set import OrderedSet

from pacman.model.resources import AbstractSDRAM, MultiRegionSDRAM
from pacman.model.partitioner_splitters import AbstractSplitterCommon
from pacman.model.graphs import AbstractEdgePartition, AbstractVertex
from pacman.model.graphs.application import ApplicationVertex
from pacman.model.graphs.machine import (
    MachineEdge, SourceSegmentedSDRAMMachinePartition, SDRAMMachineEdge,
    MulticastEdgePartition, MachineVertex)
from pacman.model.graphs.common import Slice
from pacman.utilities.utility_objs import ChipCounter

from spinn_front_end_common.utilities.constants import BYTES_PER_WORD

from spynnaker.pyNN.models.projection import Projection
from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.models.common import PopulationApplicationVertex
from spynnaker.pyNN.models.neuron import (
    PopulationNeuronsMachineVertex, PopulationSynapsesMachineVertexLead,
    PopulationSynapsesMachineVertexShared, NeuronProvenance, SynapseProvenance,
    SpikeProcessingFastProvenance)
from spynnaker.pyNN.models.neuron.population_neurons_machine_vertex import (
    SDRAM_PARAMS_SIZE as NEURONS_SDRAM_PARAMS_SIZE, NeuronMainProvenance)
from spynnaker.pyNN.models.neuron.synapse_dynamics import (
    SynapseDynamicsStatic, AbstractSynapseDynamicsStructural)
from spynnaker.pyNN.models.utility_models.delays import DelayExtensionVertex
from spynnaker.pyNN.models.neuron.synaptic_matrices import SynapticMatrices
from spynnaker.pyNN.models.neuron.neuron_data import NeuronData
from spynnaker.pyNN.models.abstract_models import SendsSynapticInputsOverSDRAM
from spynnaker.pyNN.models.neuron.population_synapses_machine_vertex_common \
    import (
        SDRAM_PARAMS_SIZE as SYNAPSES_SDRAM_PARAMS_SIZE, KEY_CONFIG_SIZE,
        PopulationSynapsesMachineVertexCommon)
from spynnaker.pyNN.models.neuron.synaptic_matrices import (
    SynapseRegionReferences)
from spynnaker.pyNN.utilities.constants import (
    SYNAPSE_SDRAM_PARTITION_ID, SPIKE_PARTITION_ID, MAX_RING_BUFFER_BITS)
from spynnaker.pyNN.models.spike_source import SpikeSourcePoissonVertex
from spynnaker.pyNN.models.neural_projections import ProjectionApplicationEdge
from spynnaker.pyNN.utilities.utility_calls import get_n_bits
from spynnaker.pyNN.exceptions import SynapticConfigurationException
from spynnaker.pyNN.models.neuron.master_pop_table import (
    MasterPopTableAsBinarySearch)
from spynnaker.pyNN.utilities.bit_field_utilities import (
    get_sdram_for_bit_field_region)
from spynnaker.pyNN.models.spike_source.spike_source_poisson_machine_vertex \
    import (
        SpikeSourcePoissonMachineVertex)

from .splitter_population_vertex import SplitterPopulationVertex
from .abstract_supports_one_to_one_sdram_input import (
    AbstractSupportsOneToOneSDRAMInput)

logger = FormatAdapter(logging.getLogger(__name__))


class SplitterPopulationVertexNeuronsSynapses(
        SplitterPopulationVertex, AbstractSupportsOneToOneSDRAMInput):
    """
    Splits an :py:class:`PopulationVertex` so that there are separate
    neuron cores each being fed by one or more synapse cores.  Incoming
    one-to-one Poisson cores are also added here if they meet the criteria.
    """

    __slots__ = (
        # All the neuron cores
        "__neuron_vertices",
        # All the synapse cores
        "__synapse_vertices",
        # Any application Poisson sources that are handled here
        "__poisson_sources",
        # The next synapse core to use for an incoming machine edge
        "__next_synapse_index",
        # The incoming vertices cached
        "__incoming_vertices",
        # The internal multicast partitions
        "__multicast_partitions",
        # The internal SDRAM partitions
        "__sdram_partitions",
        # The same chip groups
        "__same_chip_groups",
        # The application vertex sources that are neuromodulators
        "__neuromodulators")

    def __init__(self) -> None:
        super().__init__()

        self.__next_synapse_index = 0
        self.__neuron_vertices: List[PopulationNeuronsMachineVertex] = []
        self.__synapse_vertices: List[
            PopulationSynapsesMachineVertexCommon] = []
        self.__multicast_partitions: List[MulticastEdgePartition] = []
        self.__sdram_partitions: List[
            SourceSegmentedSDRAMMachinePartition] = []
        self.__same_chip_groups: List[Tuple[
            List[MachineVertex], AbstractSDRAM]] = []
        self.__neuromodulators: Set[ApplicationVertex] = set()
        self.__incoming_vertices: List[List[MachineVertex]] = []
        self.__poisson_sources: Set[SpikeSourcePoissonVertex] = set()

    @overrides(AbstractSplitterCommon.create_machine_vertices)
    def create_machine_vertices(self, chip_counter: ChipCounter) -> None:
        label = self.governed_app_vertex.label or ""

        # Structural plasticity can only be run on a single synapse core
        if (isinstance(self.governed_app_vertex.synapse_dynamics,
                       AbstractSynapseDynamicsStructural) and
                self.governed_app_vertex.n_synapse_cores_required != 1):
            raise SynapticConfigurationException(
                "The current implementation of structural plasticity can only"
                " be run on a single synapse core.  Please ensure the number"
                " of synapse cores is set to 1")

        # Do some checks to make sure everything is likely to fit
        n_atom_bits = self.governed_app_vertex.get_n_atom_bits()
        n_synapse_types = \
            self.governed_app_vertex.neuron_impl.get_n_synapse_types()
        if (n_atom_bits + get_n_bits(n_synapse_types) +
                get_n_bits(self.max_support_delay())) > MAX_RING_BUFFER_BITS:
            raise SynapticConfigurationException(
                "The combination of the number of neurons per core "
                f"({n_atom_bits}), the number of synapse types "
                f"({n_synapse_types}), and the maximum delay per core "
                f"({self.max_support_delay()}) will require too much DTCM. "
                "Please reduce one or more of these values.")

        incoming_direct_poisson = self.__handle_poisson_sources(label)

        atoms_per_core = min(
            self.governed_app_vertex.get_max_atoms_per_core(),
            self.governed_app_vertex.n_atoms)

        # Work out the ring buffer shifts based on all incoming things
        rb_shifts = self.governed_app_vertex.get_ring_buffer_shifts()
        weight_scales = self.governed_app_vertex.get_weight_scales(rb_shifts)

        # We add the SDRAM edge SDRAM to the neuron resources so it is
        # accounted for within the placement
        n_synapse_cores = self.governed_app_vertex.n_synapse_cores_required
        n_incoming = n_synapse_cores + len(self.__poisson_sources)
        edge_sdram = PopulationNeuronsMachineVertex.get_n_bytes_for_transfer(
            atoms_per_core, n_synapse_types)
        sdram_edge_sdram = edge_sdram * n_incoming

        # Get maximum resources for neurons for each split
        neuron_sdram = self.__get_neuron_sdram(
            atoms_per_core, sdram_edge_sdram)

        # Get resources for synapses
        structural_sz = max(
            self.governed_app_vertex.get_structural_dynamics_size(
                atoms_per_core), BYTES_PER_WORD)
        all_syn_block_sz = max(
            self.governed_app_vertex.get_synapses_size(atoms_per_core),
            BYTES_PER_WORD)
        shared_synapse_sdram = self.__get_shared_synapse_sdram(
            atoms_per_core, all_syn_block_sz, structural_sz)
        lead_synapse_core_sdram = self.__get_synapse_sdram(
            atoms_per_core, shared_synapse_sdram)
        shared_synapse_core_sdram = self.__get_synapse_sdram(atoms_per_core)
        synapse_regions = PopulationSynapsesMachineVertexLead.SYNAPSE_REGIONS
        synaptic_matrices = SynapticMatrices(
            self.governed_app_vertex, synapse_regions, atoms_per_core,
            weight_scales, all_syn_block_sz)
        neuron_data = NeuronData(self.governed_app_vertex)

        for index in range(n_synapse_cores):
            self.__incoming_vertices.append([])

        for index, vertex_slice in enumerate(self._get_fixed_slices()):
            # Create the neuron vertex for the slice
            neuron_vertex = self.__add_neuron_core(
                vertex_slice, neuron_sdram, label, index, rb_shifts,
                weight_scales, neuron_data, atoms_per_core)
            sdram: AbstractSDRAM = neuron_sdram
            source_vertices: List[MachineVertex] = list()
            source_sdram_vertices: List[SendsSynapticInputsOverSDRAM] = list()

            # Add the first vertex
            synapse_references, syn_label, feedback_partition, lead_vtx = \
                self.__add_lead_synapse_core(
                    vertex_slice, structural_sz, lead_synapse_core_sdram,
                    label, rb_shifts, weight_scales, neuron_vertex,
                    atoms_per_core, synaptic_matrices)
            sdram += lead_synapse_core_sdram
            source_vertices.append(lead_vtx)
            source_sdram_vertices.append(lead_vtx)

            # Do the remaining synapse cores
            for i in range(1, n_synapse_cores):
                shared_vtx = self.__add_shared_synapse_core(
                    syn_label, i, vertex_slice, synapse_references,
                    shared_synapse_core_sdram, feedback_partition,
                    neuron_vertex)
                sdram += shared_synapse_core_sdram
                source_vertices.append(shared_vtx)
                source_sdram_vertices.append(shared_vtx)

            # Add resources for Poisson vertices
            poisson_vertices = incoming_direct_poisson[vertex_slice]
            for poisson_vertex, _possion_edge in poisson_vertices:
                source_vertices.append(poisson_vertex)
                source_sdram_vertices.append(poisson_vertex)
                sdram += poisson_vertex.sdram_required

            # Add the cores
            n_cores = n_incoming + 1
            chip_counter.add_core(sdram, n_cores=n_cores)

            # Create an SDRAM edge partition
            sdram_partition = SourceSegmentedSDRAMMachinePartition(
                SYNAPSE_SDRAM_PARTITION_ID, source_vertices)
            self.__sdram_partitions.append(sdram_partition)
            neuron_vertex.set_sdram_partition(sdram_partition)

            # Add SDRAM edges for synapse vertices
            for source_vertex in source_vertices:
                sdram_partition.add_edge(SDRAMMachineEdge(
                    source_vertex, neuron_vertex,
                    f"SDRAM {source_vertex.label}-->{neuron_vertex.label}"))
            for source_vertex_over_sdram in source_sdram_vertices:
                source_vertex_over_sdram.set_sdram_partition(sdram_partition)

            self.__same_chip_groups.append(
                ([*source_vertices, neuron_vertex], sdram))

        # Find incoming neuromodulators
        for proj in self.governed_app_vertex.incoming_projections:
            # pylint: disable=protected-access
            edge = proj._projection_edge
            if edge.is_neuromodulation:
                self.__neuromodulators.add(edge.pre_vertex)

    def __add_neuron_core(
            self, vertex_slice: Slice, sdram: AbstractSDRAM,
            label: str, index: int, rb_shifts: List[int],
            weight_scales: NDArray[floating], neuron_data: NeuronData,
            atoms_per_core: int) -> PopulationNeuronsMachineVertex:
        """
        Add a neuron core for for a slice of neurons.

        :param vertex_slice:
            The slice of neurons to put on the core
        :param sdram:
        :param label: The name to give the core
        :param index: The index of the slice in the ordered list of slices
        :param rb_shifts:
            The computed ring-buffer shift values to use to get the weights
            back to S1615 values
        :param weight_scales:
            The scale to apply to weights to encode them in the 16-bit synapses
        :return: The neuron vertex created and the resources used
        """
        neuron_vertex = PopulationNeuronsMachineVertex(
            sdram,
            f"{label}_Neurons:{vertex_slice.lo_atom}",
            self.governed_app_vertex, vertex_slice, index, rb_shifts,
            weight_scales, neuron_data, atoms_per_core)
        self.governed_app_vertex.remember_machine_vertex(neuron_vertex)
        self.__neuron_vertices.append(neuron_vertex)

        return neuron_vertex

    def __add_lead_synapse_core(
            self, vertex_slice: Slice, structural_sz: int,
            lead_synapse_core_sdram: AbstractSDRAM, label: str,
            rb_shifts: List[int], weight_scales: NDArray[floating],
            neuron_vertex: PopulationNeuronsMachineVertex,
            atoms_per_core: int, synaptic_matrices: SynapticMatrices) -> Tuple[
                SynapseRegionReferences, str,
                Optional[MulticastEdgePartition],
                PopulationSynapsesMachineVertexLead]:
        """
        Add the first synapse core for a neuron core.  This core will
        generate all the synaptic data required.

        :param vertex_slice:
            The slice of neurons on the neuron core
        :param lead_synapse_core_sdram:
            The SDRAM that will be used by every lead synapse core
        :param label: The name to give the core
        :param rb_shifts:
            The computed ring-buffer shift values to use to get the weights
            back to S1615 values
        :param weight_scales:
            The scale to apply to weights to encode them in the 16-bit synapses
        :param neuron_vertex:
            The neuron vertex the synapses will feed into
        :param atoms_per_core: The maximum atoms per core
        :return: References to the synapse regions that can be used by a shared
            synapse core, the basic label for the synapse cores,
            the feedback partition (if needed), and the lead synapse core
        """
        synapse_references = SynapseRegionReferences(
            *SpynnakerDataView.get_next_ds_references(7))
        syn_label = (
            f"{label}_Synapses:{vertex_slice.lo_atom}")

        # Do the lead synapse core
        lead_synapse_vertex = PopulationSynapsesMachineVertexLead(
            lead_synapse_core_sdram, f"{syn_label}(0)",
            self.governed_app_vertex, vertex_slice, rb_shifts, weight_scales,
            structural_sz, synapse_references, atoms_per_core,
            synaptic_matrices)
        self.governed_app_vertex.remember_machine_vertex(lead_synapse_vertex)
        self.__synapse_vertices.append(lead_synapse_vertex)
        self.__incoming_vertices[0].append(lead_synapse_vertex)

        part = self.__add_plastic_feedback(neuron_vertex, lead_synapse_vertex)

        return synapse_references, syn_label, part, lead_synapse_vertex

    def __add_shared_synapse_core(
            self, syn_label: str, s_index: int, vertex_slice: Slice,
            synapse_references: SynapseRegionReferences,
            shared_synapse_sdram: AbstractSDRAM,
            feedback_partition: Optional[AbstractEdgePartition],
            neuron_vertex: PopulationNeuronsMachineVertex)\
            -> PopulationSynapsesMachineVertexShared:
        """
        Add a second or subsequent synapse core.  This will reference the
        synaptic data generated by the lead synapse core.

        :param syn_label: The basic synapse core label to be extended
        :param s_index: The index of the synapse core (0 is the lead core)
        :param vertex_slice: The slice of neurons on the neuron core
        :param synapse_references: References to the synapse regions
        :param shared_synapse_sdram:
        :param feedback_partition:
`       :param neuron_vertex:
            The neuron vertex the synapses will feed into
        :return: The
        """
        synapse_label = f"{syn_label}({s_index})"
        synapse_vertex = PopulationSynapsesMachineVertexShared(
            shared_synapse_sdram, synapse_label, self.governed_app_vertex,
            vertex_slice, synapse_references)
        self.governed_app_vertex.remember_machine_vertex(synapse_vertex)
        self.__synapse_vertices.append(synapse_vertex)
        self.__incoming_vertices[s_index].append(synapse_vertex)

        if feedback_partition is not None:
            neuron_to_synapse_edge = MachineEdge(neuron_vertex, synapse_vertex)
            feedback_partition.add_edge(neuron_to_synapse_edge)
            synapse_vertex.set_neuron_vertex_and_partition_id(
                neuron_vertex, SPIKE_PARTITION_ID)

        return synapse_vertex

    def __add_plastic_feedback(
            self, neuron_vertex: PopulationNeuronsMachineVertex,
            synapse_vertex: PopulationSynapsesMachineVertexCommon) -> Optional[
                MulticastEdgePartition]:
        """
        Add an edge if needed from the neuron core back to the synapse core
        to allow the synapse core to process plastic synapses.

        :param neuron_vertex: The neuron vertex to start the edge at
        :param synapse_vertex: A synapse vertex to feed the spikes back to
        """
        # If synapse dynamics is not simply static, link the neuron vertex
        # back to the synapse vertex
        if self.governed_app_vertex.synapse_dynamics is not None and \
                not isinstance(self.governed_app_vertex.synapse_dynamics,
                               SynapseDynamicsStatic):
            if self.governed_app_vertex.self_projection is None:
                feedback_partition = MulticastEdgePartition(
                    neuron_vertex, SPIKE_PARTITION_ID)
                neuron_to_synapse_edge = MachineEdge(
                    neuron_vertex, synapse_vertex)
                feedback_partition.add_edge(neuron_to_synapse_edge)
                self.__multicast_partitions.append(feedback_partition)
                synapse_vertex.set_neuron_vertex_and_partition_id(
                    neuron_vertex, SPIKE_PARTITION_ID)
                return feedback_partition
            synapse_vertex.set_neuron_vertex_and_partition_id(
                neuron_vertex, SPIKE_PARTITION_ID)
        return None

    @property
    def __too_many_cores(self) -> bool:
        version = SpynnakerDataView.get_machine_version()
        incoming = self.governed_app_vertex.incoming_poisson_projections
        n_synapse_cores = self.governed_app_vertex.n_synapse_cores_required
        return ((len(incoming) + n_synapse_cores + 1) >=
                (version.max_cores_per_chip - version.n_scamp_cores))

    def __handle_poisson_sources(self, label: str) -> Dict[Slice, List[Tuple[
            SpikeSourcePoissonMachineVertex, ProjectionApplicationEdge]]]:
        """
        Go through the incoming projections and find Poisson sources with
        splitters that work with us, and one-to-one connections that will
        then work with SDRAM.

        :param label: Base label to give to the Poisson cores
        """
        # The only way to avoid circular imports is to import here
        # pylint: disable=import-outside-toplevel
        from spynnaker.pyNN.extra_algorithms.splitter_components\
            .splitter_utils import is_direct_poisson_source
        incoming_direct_poisson: Dict[Slice, List[Tuple[
            SpikeSourcePoissonMachineVertex,
            ProjectionApplicationEdge]]] = defaultdict(list)
        # If there are going to be too many to fit on a chip, don't do direct
        # Poisson
        if self.__too_many_cores:
            return incoming_direct_poisson
        for proj in self.governed_app_vertex.incoming_poisson_projections:
            # pylint: disable=protected-access
            edge = proj._projection_edge
            pre_vertex = cast(SpikeSourcePoissonVertex, edge.pre_vertex)
            conn = proj._synapse_information.connector
            dynamics = proj._synapse_information.synapse_dynamics
            delay = proj._synapse_information.delays
            if is_direct_poisson_source(self.governed_app_vertex, pre_vertex,
                                        conn, dynamics, delay):
                # Create the direct Poisson vertices here; the splitter
                # for the Poisson will create any others as needed
                for vertex_slice in self._get_fixed_slices():
                    sdram = pre_vertex.get_sdram_used_by_atoms(vertex_slice)
                    poisson_m_vertex = pre_vertex.create_machine_vertex(
                        vertex_slice, sdram, label=(
                            f"{label}_Poisson:"
                            f"{vertex_slice.lo_atom}"))
                    pre_vertex.remember_machine_vertex(poisson_m_vertex)
                    incoming_direct_poisson[vertex_slice].append(
                        (poisson_m_vertex, edge))

                # Keep track of sources that have been handled
                self.__poisson_sources.add(pre_vertex)
        return incoming_direct_poisson

    @overrides(AbstractSupportsOneToOneSDRAMInput.handles_source_vertex)
    def handles_source_vertex(self, projection: Projection) -> bool:
        # If there are too many incoming Poisson sources, we can't do this
        if self.__too_many_cores:
            return False

        # The only way to avoid circular imports is to import here
        # pylint: disable=import-outside-toplevel
        from spynnaker.pyNN.extra_algorithms.splitter_components\
            .splitter_utils import is_direct_poisson_source

        # pylint: disable=protected-access
        edge = projection._projection_edge
        pre_vertex = edge.pre_vertex
        connector = projection._synapse_information.connector
        dynamics = projection._synapse_information.synapse_dynamics
        delay = projection._synapse_information.delays
        return is_direct_poisson_source(
            self.governed_app_vertex, pre_vertex, connector, dynamics, delay)

    @overrides(AbstractSplitterCommon.get_in_coming_slices)
    def get_in_coming_slices(self) -> Sequence[Slice]:
        return self._get_fixed_slices()

    @overrides(AbstractSplitterCommon.get_out_going_slices)
    def get_out_going_slices(self) -> Sequence[Slice]:
        return self._get_fixed_slices()

    @overrides(AbstractSplitterCommon.get_out_going_vertices)
    def get_out_going_vertices(self, partition_id: str) -> Sequence[
            PopulationNeuronsMachineVertex]:
        return self.__neuron_vertices

    @overrides(AbstractSplitterCommon.get_in_coming_vertices)
    def get_in_coming_vertices(self, partition_id: str) -> Sequence[
            PopulationSynapsesMachineVertexCommon]:
        return self.__synapse_vertices

    @overrides(AbstractSplitterCommon.get_source_specific_in_coming_vertices)
    def get_source_specific_in_coming_vertices(
            self, source_vertex: ApplicationVertex,
            partition_id: str) -> Sequence[
                Tuple[MachineVertex, Sequence[AbstractVertex]]]:
        # If delayed get the real pre-vertex
        if isinstance(source_vertex, DelayExtensionVertex):
            pre_vertex = cast(
                PopulationApplicationVertex, source_vertex.source_vertex)
        elif isinstance(source_vertex, PopulationApplicationVertex):
            pre_vertex = source_vertex
        else:
            raise ValueError(
                f"unsupported source vertex type: {type(source_vertex)}")

        # Filter out edges from Poisson sources being done using SDRAM
        if pre_vertex in self.__poisson_sources:
            return []

        # If the incoming edge targets the reward or punishment receptors
        # then it needs to be treated differently
        if pre_vertex in self.__neuromodulators:
            # In this instance, choose to send to all synapse vertices
            return [(v, [source_vertex])
                    for s in self.__incoming_vertices for v in s]

        # Get the set of connected sources overall using the real pre-vertex
        targets: Dict[MachineVertex, OrderedSet[AbstractVertex]] = defaultdict(
            OrderedSet)
        for proj in self.governed_app_vertex.get_incoming_projections_from(
                pre_vertex):
            # pylint: disable=protected-access
            s_info = proj._synapse_information
            # Use the original source vertex here to ensure the actual machine
            # vertices of the source vertex make it in
            for (tgt, srcs) in s_info.synapse_dynamics.get_connected_vertices(
                    s_info, source_vertex, self.governed_app_vertex):
                targets[tgt].update(srcs)

        # Split the incoming machine vertices so that they are in ~power of 2
        # groups, using the original source vertex to get the right machine
        # vertices
        sources = source_vertex.splitter.get_out_going_vertices(partition_id)
        n_sources = len(sources)
        n_synapse_cores = self.governed_app_vertex.n_synapse_cores_required
        sources_per_vertex = max(1, int(2 ** math.ceil(math.log2(
            n_sources / n_synapse_cores))))

        # Start on a different index each time to "even things out"
        index = self.__next_synapse_index
        self.__next_synapse_index = (
            (self.__next_synapse_index + 1) % n_synapse_cores)
        result: List[Tuple[MachineVertex, List[MachineVertex]]] = list()
        for start in range(0, n_sources, sources_per_vertex):
            end = min(start + sources_per_vertex, n_sources)
            source_range = sources[start:end]
            for s_vertex in self.__incoming_vertices[index]:
                targets_filtered = targets[s_vertex]
                filtered = [s for s in source_range
                            if (s in targets_filtered or
                                s.app_vertex in targets_filtered)]
                result.append((s_vertex, filtered))
            index = (index + 1) % n_synapse_cores

        return result

    @overrides(AbstractSplitterCommon.machine_vertices_for_recording)
    def machine_vertices_for_recording(
            self, variable_to_record: str) -> Sequence[MachineVertex]:
        if self.governed_app_vertex.neuron_recorder.is_recordable(
                variable_to_record):
            return self.__neuron_vertices
        return self.__synapse_vertices

    @overrides(AbstractSplitterCommon.reset_called)
    def reset_called(self) -> None:
        self.__neuron_vertices = []
        self.__synapse_vertices = []
        self.__multicast_partitions = []
        self.__sdram_partitions = []
        self.__same_chip_groups = []
        self.__poisson_sources = set()
        self.__incoming_vertices = []

    def __get_neuron_sdram(
            self, n_atoms: int, sdram_edge_sdram: int) -> MultiRegionSDRAM:
        """
        Gets the resources of the neurons of a slice of atoms from a given
        application vertex.
        """
        n_record = len(self.governed_app_vertex.neuron_recordables)
        variable_sdram = \
            self.governed_app_vertex.get_max_neuron_variable_sdram(n_atoms)
        sdram = MultiRegionSDRAM()
        sdram.merge(self.governed_app_vertex.get_common_constant_sdram(
            n_record, NeuronProvenance.N_ITEMS + NeuronMainProvenance.N_ITEMS,
            PopulationNeuronsMachineVertex.COMMON_REGIONS))
        sdram.merge(self.governed_app_vertex.get_neuron_constant_sdram(
            n_atoms, PopulationNeuronsMachineVertex.NEURON_REGIONS))
        sdram.add_cost(
            PopulationNeuronsMachineVertex.REGIONS.SDRAM_EDGE_PARAMS,
            NEURONS_SDRAM_PARAMS_SIZE)
        sdram.nest(
            len(PopulationNeuronsMachineVertex.REGIONS), variable_sdram)
        sdram.add_cost(
            len(PopulationNeuronsMachineVertex.REGIONS) + 1, sdram_edge_sdram)

        # return the total resources.
        return sdram

    def __shared_synapse_sdram(
            self, independent_synapse_sdram: MultiRegionSDRAM,
            proj_dependent_sdram: MultiRegionSDRAM, all_syn_block_sz: int,
            structural_sz: int, dynamics_sz: int) -> MultiRegionSDRAM:
        """
        Get the SDRAM shared between synapse cores.
        """
        regions = PopulationSynapsesMachineVertexLead.SYNAPSE_REGIONS
        sdram = MultiRegionSDRAM()
        sdram.merge(independent_synapse_sdram)
        sdram.merge(proj_dependent_sdram)
        sdram.add_cost(regions.synaptic_matrix, all_syn_block_sz)
        sdram.add_cost(regions.structural_dynamics, structural_sz)
        sdram.add_cost(regions.synapse_dynamics, dynamics_sz)
        return sdram

    def __get_shared_synapse_sdram(
            self, n_atoms: int, all_syn_block_sz: int,
            structural_sz: int) -> MultiRegionSDRAM:
        independent_synapse_sdram = self.__independent_synapse_sdram()
        proj_dependent_sdram = self.__proj_dependent_synapse_sdram()
        dynamics_sz = self.governed_app_vertex.get_synapse_dynamics_size(
            n_atoms)
        dynamics_sz = max(dynamics_sz, BYTES_PER_WORD)
        return self.__shared_synapse_sdram(
            independent_synapse_sdram, proj_dependent_sdram,
            all_syn_block_sz, structural_sz, dynamics_sz)

    def __get_synapse_sdram(
            self, n_atoms: int,
            shared_sdram: Optional[MultiRegionSDRAM] = None
            ) -> MultiRegionSDRAM:
        """
        Get the resources of the synapses of a slice of atoms from a
        given application vertex.

        :param shared_sdram:
            The SDRAM shared between cores, if this is to be included
        """
        n_record = len(self.governed_app_vertex.synapse_recordables)
        variable_sdram = \
            self.governed_app_vertex.get_max_synapse_variable_sdram(n_atoms)
        sdram = MultiRegionSDRAM()
        sdram.merge(self.governed_app_vertex.get_common_constant_sdram(
            n_record,
            SynapseProvenance.N_ITEMS + SpikeProcessingFastProvenance.N_ITEMS,
            PopulationSynapsesMachineVertexLead.COMMON_REGIONS))

        sdram.add_cost(
            PopulationSynapsesMachineVertexLead.REGIONS.SDRAM_EDGE_PARAMS,
            SYNAPSES_SDRAM_PARAMS_SIZE)
        sdram.add_cost(
            PopulationSynapsesMachineVertexLead.REGIONS.KEY_REGION,
            KEY_CONFIG_SIZE)
        sdram.nest(
            len(PopulationSynapsesMachineVertexLead.REGIONS) + 1,
            variable_sdram)
        if shared_sdram is not None:
            sdram.merge(shared_sdram)

        # return the total resources.
        return sdram

    def __independent_synapse_sdram(self) -> MultiRegionSDRAM:
        """
        Get the SDRAM used by all synapse cores independent of projections.
        """
        regions = PopulationSynapsesMachineVertexLead.SYNAPSE_REGIONS
        sdram = MultiRegionSDRAM()
        sdram.add_cost(
            regions.synapse_params,
            max(self.governed_app_vertex.get_synapse_params_size(),
                BYTES_PER_WORD))
        return sdram

    def __proj_dependent_synapse_sdram(self) -> MultiRegionSDRAM:
        """
        Get the SDRAM used by synapse cores dependent on the projections.
        """
        regions = PopulationSynapsesMachineVertexLead.SYNAPSE_REGIONS
        sdram = MultiRegionSDRAM()
        sdram.add_cost(
            regions.pop_table,
            max(MasterPopTableAsBinarySearch.get_master_population_table_size(
                self.governed_app_vertex.incoming_projections),
                BYTES_PER_WORD))
        sdram.add_cost(
            regions.connection_builder,
            max(self.governed_app_vertex.get_synapse_expander_size(),
                BYTES_PER_WORD))
        sdram.add_cost(
            regions.bitfield_filter,
            max(get_sdram_for_bit_field_region(
                self.governed_app_vertex.incoming_projections),
                BYTES_PER_WORD))
        return sdram

    @overrides(AbstractSplitterCommon.get_same_chip_groups)
    def get_same_chip_groups(self) -> List[
            Tuple[List[MachineVertex], AbstractSDRAM]]:
        return self.__same_chip_groups

    @overrides(AbstractSplitterCommon.get_internal_multicast_partitions)
    def get_internal_multicast_partitions(
            self) -> List[MulticastEdgePartition]:
        return self.__multicast_partitions

    @overrides(AbstractSplitterCommon.get_internal_sdram_partitions)
    def get_internal_sdram_partitions(
            self) -> List[SourceSegmentedSDRAMMachinePartition]:
        return self.__sdram_partitions



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import List, Optional, Tuple, TYPE_CHECKING, Sequence

import numpy
from numpy.typing import NDArray

from spinn_utilities.overrides import overrides

from pacman.model.graphs.common import Slice

from .abstract_connector import AbstractConnector
from .abstract_generate_connector_on_host import (
    AbstractGenerateConnectorOnHost)
try:
    import csa  # type: ignore[import]
    _csa_import_error: Optional[ImportError] = None
except ImportError as __ex:
    # Importing csa causes problems with readthedocs so allowing it to fail
    _csa_import_error = __ex
if TYPE_CHECKING:
    from csa.connset import CSet  # type: ignore[import]
    from spynnaker.pyNN.models.neural_projections import SynapseInformation


class CSAConnector(AbstractConnector, AbstractGenerateConnectorOnHost):
    """
    Make connections using a Connection Set Algebra (Djurfeldt 2012)
    description between the neurons in the pre- and post-populations.

    .. note::
        If you get TypeError in Python 3 see:
        https://github.com/INCF/csa/issues/10
    """

    __slots__ = (
        "__cset",
        "__full_connection_set",
        "__full_cset")

    def __init__(self, cset: CSet, safe: bool = True, callback: None = None,
                 verbose: bool = False):
        """
        :param cset:
            A description of the connection set between populations
        :param safe:
            If ``True``, check that weights and delays have valid values.
            If ``False``, this check is skipped.
        :param callback:
            if given, a callable that display a progress bar on the terminal.

            .. note::
                Not supported by sPyNNaker.

        :param verbose:
            Whether to output extra information about the connectivity to a
            CSV file
        :raises ImportError:
            if the `csa` library isn't present; it's tricky to install in
            some environments so we don't force it to be present unless you
            want to actually use this class.
        """
        super().__init__(safe, callback, verbose)
        if _csa_import_error:
            raise _csa_import_error
        self.__cset = cset

        # Storage for full connection sets
        self.__full_connection_set: Optional[List[CSet]] = None
        self.__full_cset: Optional[List[CSet]] = None

    @overrides(AbstractConnector.get_delay_maximum)
    def get_delay_maximum(self, synapse_info: SynapseInformation) -> float:
        n_conns_max = synapse_info.n_pre_neurons * synapse_info.n_post_neurons
        # we can probably look at the array and do better than this?
        return self._get_delay_maximum(
            synapse_info.delays, n_conns_max, synapse_info)

    @overrides(AbstractConnector.get_delay_minimum)
    def get_delay_minimum(self, synapse_info: SynapseInformation) -> float:
        n_conns_max = synapse_info.n_pre_neurons * synapse_info.n_post_neurons
        # we can probably look at the array and do better than this?
        return self._get_delay_minimum(
            synapse_info.delays, n_conns_max, synapse_info)

    def _get_n_connections(
            self, post_vertex_slice: Slice,
            synapse_info: SynapseInformation) -> Tuple[int, CSet]:
        # do the work from self._cset in here

        # this is where the magic needs to happen somehow
        if self.__full_cset is None:
            self.__full_cset = [x for x in csa.cross(
                range(synapse_info.n_pre_neurons),
                range(synapse_info.n_post_neurons)) * self.__cset]

        # use CSA to cross the range of this vertex's neurons with the cset
        pair_list = (
            csa.cross(
                range(synapse_info.n_pre_neurons),
                list(int(x) for x in post_vertex_slice.get_raster_ids()))
            * self.__full_cset)

        if self.verbose:
            print('full cset: ', self.__full_cset)
            print('this vertex pair_list: ', pair_list)
            print('this vertex pre_neurons: ', [x[0] for x in pair_list])
            print('this vertex post_neurons: ', [x[1] for x in pair_list])

        n_connections = len(pair_list)  # size of the array created
        return n_connections, pair_list

    @overrides(AbstractConnector.get_n_connections_from_pre_vertex_maximum)
    def get_n_connections_from_pre_vertex_maximum(
            self, n_post_atoms: int, synapse_info: SynapseInformation,
            min_delay: Optional[float] = None,
            max_delay: Optional[float] = None) -> int:
        if min_delay is None or max_delay is None:
            raise ValueError("min_delay and max_delay must be supplied")
        n_connections_max = n_post_atoms
        return self._get_n_connections_from_pre_vertex_with_delay_maximum(
            synapse_info.delays,
            synapse_info.n_pre_neurons * synapse_info.n_post_neurons,
            n_connections_max, min_delay, max_delay, synapse_info)

    @overrides(AbstractConnector.get_n_connections_to_post_vertex_maximum)
    def get_n_connections_to_post_vertex_maximum(
            self, synapse_info: SynapseInformation) -> int:
        n_connections_max = synapse_info.n_pre_neurons
        return n_connections_max

    @overrides(AbstractConnector.get_weight_maximum)
    def get_weight_maximum(self, synapse_info: SynapseInformation) -> float:
        n_conns_max = synapse_info.n_pre_neurons * synapse_info.n_post_neurons
        return self._get_weight_maximum(
            synapse_info.weights, n_conns_max, synapse_info)

    @overrides(AbstractGenerateConnectorOnHost.create_synaptic_block)
    def create_synaptic_block(
            self, post_slices: Sequence[Slice], post_vertex_slice: Slice,
            synapse_type: int, synapse_info: SynapseInformation) -> NDArray:
        n_connections, pair_list = self._get_n_connections(
            post_vertex_slice, synapse_info)

        # Use whatever has been set up in _get_n_connections here
        # to send into the block structure
        # TO DO: not sure this works with repeated connector, but
        #        what does a repeated connector mean in this context?
        if self.__full_connection_set is None:
            self.__full_connection_set = [x for x in pair_list]
        else:
            self.__full_connection_set += [x for x in pair_list]

        block = numpy.zeros(
            n_connections, dtype=AbstractConnector.NUMPY_SYNAPSES_DTYPE)
        # source and target are the pre_neurons and post_neurons in pair_list
        block["source"] = synapse_info.pre_vertex.get_key_ordered_indices(
            numpy.array([x[0] for x in pair_list]))
        block["target"] = post_vertex_slice.get_relative_indices(
            numpy.array([x[1] for x in pair_list]))
        block["weight"] = self._generate_weights(
            block["source"], block["target"], n_connections, post_vertex_slice,
            synapse_info)
        block["delay"] = self._generate_delays(
            block["source"], block["target"], n_connections, post_vertex_slice,
            synapse_info)
        block["synapse_type"] = synapse_type
        return block

    def show_connection_set(
            self, n_pre_neurons: int, n_post_neurons: int) -> None:
        """
        Calls csa.show

        :param n_pre_neurons:
        :param n_post_neurons:
        """
        # Yuck; this was supposed to be available to the user from scripts...
        csa.show(self.__full_connection_set, n_pre_neurons, n_post_neurons)

    def __repr__(self) -> str:
        return f"CSAConnector({self.__full_cset})"



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import math
from typing import Optional, Sequence, TYPE_CHECKING

import numpy
from numpy.typing import NDArray

from pyNN.random import NumpyRNG

from spinn_utilities.overrides import overrides

from pacman.model.graphs.common import Slice

from spinn_front_end_common.utilities.exceptions import ConfigurationException

from .abstract_connector import AbstractConnector
from .abstract_generate_connector_on_host import (
    AbstractGenerateConnectorOnHost)

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections import SynapseInformation


class SmallWorldConnector(AbstractConnector, AbstractGenerateConnectorOnHost):
    """
    A connector that uses connection statistics based on the Small World
    network connectivity model.

    .. note::
        This is typically used from a population to itself.
    """
    __slots__ = (
        "__degree",
        "__mask",
        "__n_connections",
        "__rewiring",
        "__rng")

    def __init__(
            self, degree: float, rewiring: float,
            allow_self_connections: bool = True,
            n_connections: Optional[int] = None,
            rng: Optional[NumpyRNG] = None,
            safe: bool = True, callback: None = None, verbose: bool = False):
        """
        :param degree:
            the region length where nodes will be connected locally
        :param rewiring: the probability of rewiring each edge
        :param allow_self_connections:
            if the connector is used to connect a Population to itself, this
            flag determines whether a neuron is allowed to connect to itself,
            or only to other neurons in the Population.
            False is currently unsupported!
        :param n_connections:
            if specified, the number of efferent synaptic connections per
            neuron
        :param rng:
            Seeded random number generator, or ``None`` to make one when
            needed.
        :param safe:
            If ``True``, check that weights and delays have valid values.
            If ``False``, this check is skipped.
        :param callback:
            if given, a callable that display a progress bar on the terminal.

            .. note::
                Not supported by sPyNNaker.
        :param verbose:
            Whether to output extra information about the connectivity to a
            CSV file
        """
        super().__init__(safe, callback, verbose)
        self.__rewiring = rewiring
        self.__degree = degree
        if not allow_self_connections:
            raise NotImplementedError(
                "disabling self connections currently not supported")
        self.__mask: Optional[NDArray] = None
        self.__n_connections = 0
        self.__rng = rng or NumpyRNG()

        if n_connections is not None:
            raise NotImplementedError(
                "n_connections is not implemented for"
                " SmallWorldConnector on this platform")

    @overrides(AbstractConnector.set_projection_information)
    def set_projection_information(
            self, synapse_info: SynapseInformation) -> None:
        super().set_projection_information(synapse_info)
        self._set_n_connections(synapse_info)

    def _set_n_connections(self, synapse_info: SynapseInformation) -> None:
        if self.space is None:
            raise ConfigurationException("a metric space is required")
        # Get the probabilities up-front for now
        # TODO: Work out how this can be done statistically
        # space.distances(...) expects N,3 array in PyNN0.7, but 3,N in PyNN0.8
        pre_positions = synapse_info.pre_population.positions
        post_positions = synapse_info.post_population.positions

        distances = self.space.distances(
            pre_positions, post_positions, False)

        # PyNN 0.8 returns a flattened (C-style) array from space.distances,
        # so the easiest thing to do here is to reshape back to the "expected"
        # PyNN 0.7 shape; otherwise later code gets confusing and difficult
        if len(distances.shape) == 1:
            d = numpy.reshape(distances, (pre_positions.shape[0],
                                          post_positions.shape[0]))
        else:
            d = distances

        self.__mask = (d < self.__degree).astype(numpy.float64)
        self.__n_connections = int(math.ceil(numpy.sum(self.__mask)))

    @overrides(AbstractConnector.get_delay_maximum)
    def get_delay_maximum(self, synapse_info: SynapseInformation) -> float:
        return self._get_delay_maximum(
            synapse_info.delays, self.__n_connections, synapse_info)

    @overrides(AbstractConnector.get_delay_minimum)
    def get_delay_minimum(self, synapse_info: SynapseInformation) -> float:
        return self._get_delay_minimum(
            synapse_info.delays, self.__n_connections, synapse_info)

    @overrides(AbstractConnector.get_n_connections_from_pre_vertex_maximum)
    def get_n_connections_from_pre_vertex_maximum(
            self, n_post_atoms: int, synapse_info: SynapseInformation,
            min_delay: Optional[float] = None,
            max_delay: Optional[float] = None) -> int:
        assert self.__mask is not None
        # Break the array into n_post_atoms units
        split_positions = numpy.arange(
            0, synapse_info.n_post_neurons, n_post_atoms)
        split_array = numpy.array_split(self.__mask, split_positions)

        # Sum the 1s in each split row
        sum_rows = [numpy.sum(s, axis=1) for s in split_array]

        # Find the maximum of the rows
        n_connections = max([x for y in sum_rows for x in y])

        if min_delay is None or max_delay is None:
            return n_connections

        return self._get_n_connections_from_pre_vertex_with_delay_maximum(
            synapse_info.delays, self.__n_connections, n_connections,
            min_delay, max_delay, synapse_info)

    @overrides(AbstractConnector.get_n_connections_to_post_vertex_maximum)
    def get_n_connections_to_post_vertex_maximum(
            self, synapse_info: SynapseInformation) -> int:
        assert self.__mask is not None
        return numpy.amax([
            numpy.sum(self.__mask[:, i]) for i in range(
                synapse_info.n_post_neurons)])

    @overrides(AbstractConnector.get_weight_maximum)
    def get_weight_maximum(self, synapse_info: SynapseInformation) -> float:
        return self._get_weight_maximum(
            synapse_info.weights, self.__n_connections, synapse_info)

    @overrides(AbstractGenerateConnectorOnHost.create_synaptic_block)
    def create_synaptic_block(
            self, post_slices: Sequence[Slice], post_vertex_slice: Slice,
            synapse_type: int, synapse_info: SynapseInformation) -> NDArray:
        if self.__mask is None:
            return numpy.zeros(0, dtype=self.NUMPY_SYNAPSES_DTYPE)
        raster_ids = post_vertex_slice.get_raster_ids()
        ids = numpy.where(self.__mask[:, raster_ids])
        n_connections = len(ids[0])

        block = numpy.zeros(n_connections, dtype=self.NUMPY_SYNAPSES_DTYPE)
        block["source"] = synapse_info.pre_vertex.get_key_ordered_indices(
            ids[0] % synapse_info.n_pre_neurons)
        block["target"] = post_vertex_slice.get_relative_indices(
            raster_ids[ids[1]])
        block["weight"] = self._generate_weights(
            block["source"], block["target"], n_connections, post_vertex_slice,
            synapse_info)
        block["delay"] = self._generate_delays(
            block["source"], block["target"], n_connections, post_vertex_slice,
            synapse_info)
        block["synapse_type"] = synapse_type

        # Re-wire some connections
        rewired = numpy.where(
            self.__rng.next(n_connections) < self.__rewiring)[0]
        block["target"][rewired] = (
            (self.__rng.next(rewired.size) * (post_vertex_slice.n_atoms - 1)) +
            post_vertex_slice.lo_atom)

        return block

    def __repr__(self) -> str:
        return ("SmallWorldConnector"
                f"(degree={self.__degree}, rewiring={self.__rewiring})")



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import List, Sequence, TYPE_CHECKING

from numpy import floating
from numpy.typing import NDArray

from spinn_utilities.overrides import overrides

from pacman.model.graphs.common import Slice
from pacman.model.resources import AbstractSDRAM
from pacman.model.placements import Placement

from spinn_front_end_common.abstract_models import (
    AbstractGeneratesDataSpecification, AbstractRewritesDataSpecification)
from spinn_front_end_common.interface.ds import (
    DataSpecificationGenerator, DataSpecificationReloader)

from .population_machine_common import PopulationMachineCommon
from .population_machine_synapses import PopulationMachineSynapses
from .population_synapses_machine_vertex_common import (
    PopulationSynapsesMachineVertexCommon)

if TYPE_CHECKING:
    from .population_vertex import PopulationVertex
    from spynnaker.pyNN.models.neuron.synaptic_matrices import (
        SynapseRegions, SynapseRegionReferences, SynapticMatrices)


class PopulationSynapsesMachineVertexLead(
        PopulationSynapsesMachineVertexCommon,
        PopulationMachineSynapses,
        AbstractGeneratesDataSpecification,
        AbstractRewritesDataSpecification):
    """
    A synaptic machine vertex that leads other synaptic machine vertices,
    writing shared areas.
    """

    __slots__ = (
        "__synaptic_matrices",
        "__ring_buffer_shifts",
        "__weight_scales",
        "__structural_sz",
        "__synapse_references",
        "__max_atoms_per_core",
        "__regenerate_data")

    def __init__(
            self, sdram: AbstractSDRAM, label: str,
            app_vertex: PopulationVertex, vertex_slice: Slice,
            ring_buffer_shifts: Sequence[int],
            weight_scales: NDArray[floating], structural_sz: int,
            synapse_references: SynapseRegionReferences,
            max_atoms_per_core: int, synaptic_matrices: SynapticMatrices):
        """
        :param sdram: The SDRAM used by the vertex
        :param label: The label of the vertex
        :param app_vertex: The associated application vertex
        :param vertex_slice: The slice of the population that this implements
        """
        super().__init__(
            sdram, label, app_vertex, vertex_slice)
        self.__ring_buffer_shifts = ring_buffer_shifts
        self.__weight_scales = weight_scales
        self.__structural_sz = structural_sz
        self.__synapse_references = synapse_references
        self.__max_atoms_per_core = max_atoms_per_core
        self.__regenerate_data = False

        # Need to do this last so that the values above can be used
        self.__synaptic_matrices = synaptic_matrices

    @property
    @overrides(PopulationMachineSynapses._synapse_regions)
    def _synapse_regions(self) -> SynapseRegions:
        return self.SYNAPSE_REGIONS

    @property
    @overrides(PopulationMachineSynapses._synaptic_matrices)
    def _synaptic_matrices(self) -> SynapticMatrices:
        return self.__synaptic_matrices

    @property
    @overrides(PopulationMachineSynapses._synapse_references)
    def _synapse_references(self) -> SynapseRegionReferences:
        return self.__synapse_references

    @property
    @overrides(PopulationMachineSynapses._max_atoms_per_core)
    def _max_atoms_per_core(self) -> int:
        return self.__max_atoms_per_core

    @overrides(PopulationMachineCommon.get_recorded_region_ids)
    def get_recorded_region_ids(self) -> List[int]:
        ids = self._pop_vertex.synapse_recorder.recorded_ids_by_slice(
            self.vertex_slice)
        return ids

    @overrides(AbstractGeneratesDataSpecification.generate_data_specification)
    def generate_data_specification(self, spec: DataSpecificationGenerator,
                                    placement: Placement) -> None:
        rec_regions = self._pop_vertex.synapse_recorder.get_region_sizes(
            self.vertex_slice)
        self._write_common_data_spec(spec, rec_regions)

        self._write_synapse_data_spec(
            spec, self.__ring_buffer_shifts,
            self.__weight_scales, self.__structural_sz)

        # Write information about SDRAM
        self._write_sdram_edge_spec(spec)

        # Write information about keys
        self._write_key_spec(spec)

        # End the writing of this specification:
        spec.end_specification()

    @overrides(PopulationSynapsesMachineVertexCommon._parse_synapse_provenance)
    def _parse_synapse_provenance(
            self, label: str, x: int, y: int, p: int,
            provenance_data: Sequence[int]) -> None:
        PopulationMachineSynapses._parse_synapse_provenance(
            self, label, x, y, p, provenance_data)

    @overrides(AbstractRewritesDataSpecification.regenerate_data_specification)
    def regenerate_data_specification(self, spec: DataSpecificationReloader,
                                      placement: Placement) -> None:
        # We don't need to do anything here because the originally written
        # data can be used again
        pass

    @overrides(AbstractRewritesDataSpecification.reload_required)
    def reload_required(self) -> bool:
        return self.__regenerate_data

    @overrides(AbstractRewritesDataSpecification.set_reload_required)
    def set_reload_required(self, new_value: bool) -> None:
        self.__regenerate_data = new_value

    @overrides(PopulationMachineSynapses.set_do_synapse_regeneration)
    def set_do_synapse_regeneration(self) -> None:
        self.__regenerate_data = True



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_front_end_common.interface.ds import DataSpecificationBase
from spynnaker.pyNN.models.neuron.synapse_dynamics import (
    AbstractHasParameterNames)


class AbstractPartnerSelection(
        AbstractHasParameterNames, metaclass=AbstractBase):
    """
    A partner selection rule.
    """

    __slots__ = ()

    @property
    @abstractmethod
    def vertex_executable_suffix(self) -> str:
        """
        The suffix to be appended to the vertex executable for this rule.
        """
        raise NotImplementedError

    @abstractmethod
    def get_parameters_sdram_usage_in_bytes(self) -> int:
        """
        :returns: The amount of SDRAM used by the parameters of this rule.
        """
        raise NotImplementedError

    @abstractmethod
    def write_parameters(self, spec: DataSpecificationBase) -> None:
        """
        Write the parameters of the rule to the spec.

        :param spec:
        """
        raise NotImplementedError



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import math
from typing import Any, Iterable, List, Optional, Tuple, TYPE_CHECKING

import numpy
from numpy import floating, integer, uint8, uint16, uint32
from numpy.typing import NDArray

from pyNN.standardmodels.synapses import StaticSynapse

from spinn_utilities.overrides import overrides

from spinn_front_end_common.interface.ds import DataSpecificationBase
from spinn_front_end_common.utilities.constants import (
    BYTES_PER_WORD, BYTES_PER_SHORT)

from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.exceptions import (
    SynapticConfigurationException, InvalidParameterType)
from spynnaker.pyNN.models.neural_projections.connectors import (
    AbstractConnector)
from spynnaker.pyNN.models.neuron.plasticity.stdp.weight_dependence.\
    abstract_has_a_plus_a_minus import AbstractHasAPlusAMinus
from spynnaker.pyNN.types import WEIGHTS
from spynnaker.pyNN.types import WEIGHTS_DELAYS_IN as _In_Types
from spynnaker.pyNN.utilities.utility_calls import get_n_bits
from spynnaker.pyNN.models.neuron.synapse_dynamics.types import (
    NUMPY_CONNECTORS_DTYPE)

from .abstract_plastic_synapse_dynamics import AbstractPlasticSynapseDynamics
from .abstract_synapse_dynamics_structural import (
    AbstractSynapseDynamicsStructural)
from .abstract_generate_on_machine import (
    AbstractGenerateOnMachine, MatrixGeneratorID)
from .synapse_dynamics_neuromodulation import SynapseDynamicsNeuromodulation
from .synapse_dynamics_weight_changable import SynapseDynamicsWeightChangable
from .synapse_dynamics_weight_changer import SynapseDynamicsWeightChanger

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge, SynapseInformation)
    from spynnaker.pyNN.models.neuron.synapse_dynamics.types import (
        ConnectionsArray)
    from spynnaker.pyNN.models.neuron.plasticity.stdp.timing_dependence.\
        abstract_timing_dependence import AbstractTimingDependence
    from spynnaker.pyNN.models.neuron.plasticity.stdp.weight_dependence.\
        abstract_weight_dependence import AbstractWeightDependence
    from spynnaker.pyNN.models.neuron.synapse_io import MaxRowInfo
    from .abstract_synapse_dynamics import AbstractSynapseDynamics

# How large are the time-stamps stored with each event
TIME_STAMP_BYTES = BYTES_PER_WORD

# The targets of neuromodulation
NEUROMODULATION_TARGETS = {
    "reward": 0,
    "punishment": 1
}


class SynapseDynamicsSTDP(
        AbstractPlasticSynapseDynamics,
        AbstractGenerateOnMachine):
    """
    The dynamics of a synapse that changes over time using a
    Spike Timing Dependent Plasticity (STDP) rule.
    """

    __slots__ = (
        # Fraction of delay that is dendritic (instead of axonal or synaptic)
        "__dendritic_delay_fraction",
        # timing dependence to use for the STDP rule
        "__timing_dependence",
        # weight dependence to use for the STDP rule
        "__weight_dependence",
        # The neuromodulation instance if enabled
        "__neuromodulation",
        # padding to add to a synaptic row for synaptic rewiring
        "__pad_to_length",
        # Whether to use back-propagation delay or not
        "__backprop_delay")

    def __init__(
            self, timing_dependence: AbstractTimingDependence,
            weight_dependence: AbstractWeightDependence,
            voltage_dependence: None = None,
            dendritic_delay_fraction: float = 1.0,
            weight: _In_Types = StaticSynapse.default_parameters['weight'],
            delay: _In_Types = None, pad_to_length: Optional[int] = None,
            backprop_delay: bool = True):
        """
        :param timing_dependence:
        :param weight_dependence:
        :param voltage_dependence: not supported
        :param dendritic_delay_fraction: must be 1.0!
        :param weight:
        :param delay: Use ``None`` to get the simulator default minimum delay.
        :param pad_to_length:
        :param backprop_delay:
        """
        if timing_dependence is None or weight_dependence is None:
            raise NotImplementedError(
                "Both timing_dependence and weight_dependence must be "
                "specified")
        if voltage_dependence is not None:
            raise NotImplementedError(
                "Voltage dependence has not been implemented")
        super().__init__(delay=delay, weight=weight)
        self.__timing_dependence = timing_dependence
        self.__weight_dependence = weight_dependence
        # move data from timing to weight dependence; that's where we need it
        if isinstance(weight_dependence, AbstractHasAPlusAMinus):
            weight_dependence.set_a_plus_a_minus(
                timing_dependence.A_plus, timing_dependence.A_minus)
        self.__dendritic_delay_fraction = float(dendritic_delay_fraction)
        self.__pad_to_length = pad_to_length
        self.__backprop_delay = backprop_delay
        self.__neuromodulation: Optional[SynapseDynamicsNeuromodulation] = None

        if self.__dendritic_delay_fraction != 1.0:
            raise NotImplementedError("All delays must be dendritic!")

    def _merge_neuromodulation(
            self, neuromodulation: SynapseDynamicsNeuromodulation) -> None:
        if self.__neuromodulation is None:
            self.__neuromodulation = neuromodulation
        elif not self.__neuromodulation.is_neuromodulation_same_as(
                neuromodulation):
            raise SynapticConfigurationException(
                "Neuromodulation must match exactly when using multiple"
                " edges to the same Population")

    @overrides(AbstractPlasticSynapseDynamics.merge)
    def merge(self, synapse_dynamics: AbstractSynapseDynamics
              ) -> AbstractSynapseDynamics:
        # If dynamics is Neuromodulation, merge with other neuromodulation,
        # and then return ourselves, as neuromodulation can't be used by
        # itself
        if isinstance(synapse_dynamics, SynapseDynamicsNeuromodulation):
            self._merge_neuromodulation(synapse_dynamics)
            return self

        if isinstance(synapse_dynamics, (SynapseDynamicsWeightChangable,
                                         SynapseDynamicsWeightChanger)):
            raise SynapticConfigurationException(
                "Weight Changer and STDP are not currently compatible")

        # If dynamics is STDP, test if same as
        if isinstance(synapse_dynamics, SynapseDynamicsSTDP):
            if not self.is_same_as(synapse_dynamics):
                raise SynapticConfigurationException(
                    "Synapse dynamics must match exactly when using multiple"
                    " edges to the same population")

            if self.__neuromodulation is not None:
                # pylint: disable=protected-access
                synapse_dynamics._merge_neuromodulation(self.__neuromodulation)

            # If STDP part matches, return the other, as it might also be
            # structural
            return synapse_dynamics

        # If dynamics is structural but not STDP (as here), merge
        # NOTE: Import here as otherwise we get a circular dependency
        # pylint: disable=import-outside-toplevel
        from .synapse_dynamics_structural_stdp import (
            SynapseDynamicsStructuralSTDP)
        if isinstance(synapse_dynamics, AbstractSynapseDynamicsStructural):
            return SynapseDynamicsStructuralSTDP(
                synapse_dynamics.partner_selection, synapse_dynamics.formation,
                synapse_dynamics.elimination,
                self.timing_dependence, self.weight_dependence,
                # voltage dependence is not supported
                None, self.dendritic_delay_fraction,
                synapse_dynamics.f_rew, synapse_dynamics.initial_weight,
                synapse_dynamics.initial_delay, synapse_dynamics.s_max,
                seed=synapse_dynamics.seed,
                backprop_delay=self.backprop_delay)

        # Otherwise, it is static or neuromodulation, so return ourselves
        return self

    @overrides(AbstractPlasticSynapseDynamics.get_value)
    def get_value(self, key: str) -> Any:
        for obj in [self.__timing_dependence, self.__weight_dependence, self]:
            if hasattr(obj, key):
                return getattr(obj, key)
        raise InvalidParameterType(
            f"Type {type(self)} does not have parameter {key}")

    @overrides(AbstractPlasticSynapseDynamics.set_value)
    def set_value(self, key: str, value: Any) -> None:
        for obj in [self.__timing_dependence, self.__weight_dependence, self]:
            if hasattr(obj, key):
                setattr(obj, key, value)
                SpynnakerDataView.set_requires_mapping()
                return
        raise InvalidParameterType(
            f"Type {type(self)} does not have parameter {key}")

    @property
    def weight_dependence(self) -> AbstractWeightDependence:
        """
        The Weight Dependence component of the synapse dynamics.
        """
        return self.__weight_dependence

    @property
    def timing_dependence(self) -> AbstractTimingDependence:
        """
        timing dependence to use for the STDP rule
        """
        return self.__timing_dependence

    @property
    def dendritic_delay_fraction(self) -> float:
        """
        Settable.
        """
        return self.__dendritic_delay_fraction

    @dendritic_delay_fraction.setter
    def dendritic_delay_fraction(self, new_value: float) -> None:
        self.__dendritic_delay_fraction = new_value

    @property
    def backprop_delay(self) -> bool:
        """
        Settable.
        """
        return self.__backprop_delay

    @backprop_delay.setter
    def backprop_delay(self, backprop_delay: bool) -> None:
        self.__backprop_delay = bool(backprop_delay)

    @property
    def neuromodulation(self) -> Optional[SynapseDynamicsNeuromodulation]:
        """
        Synapses that target a neuromodulation receptor.
        """
        return self.__neuromodulation

    @overrides(AbstractPlasticSynapseDynamics.is_same_as)
    def is_same_as(self, synapse_dynamics: AbstractSynapseDynamics) -> bool:
        if not isinstance(synapse_dynamics, SynapseDynamicsSTDP):
            return False
        return (
            self.__timing_dependence.is_same_as(
                synapse_dynamics.timing_dependence) and
            self.__weight_dependence.is_same_as(
                synapse_dynamics.weight_dependence) and
            (self.__dendritic_delay_fraction ==
             synapse_dynamics.dendritic_delay_fraction))

    def get_vertex_executable_suffix(self) -> str:
        """
        :returns:
           An executable suffix based on timing, weights and neuromodulation
        """
        # Get the suffix values for timing and weight dependence
        timing_suffix = self.__timing_dependence.vertex_executable_suffix
        weight_suffix = self.__weight_dependence.vertex_executable_suffix

        if self.__neuromodulation:
            name = (
                "_stdp_" +
                self.__neuromodulation.get_vertex_executable_suffix())
        else:
            name = "_stdp_mad_"
        name += timing_suffix + "_" + weight_suffix
        return name

    def get_parameters_sdram_usage_in_bytes(
            self, n_neurons: int, n_synapse_types: int) -> int:
        """
        :param n_neurons:
        :param n_synapse_types:
        :returns: SDRAM size in bytes of parameter data
        """
        # 32-bits for back-prop delay
        size = BYTES_PER_WORD
        size += self.__timing_dependence.get_parameters_sdram_usage_in_bytes()
        size += self.__weight_dependence.get_parameters_sdram_usage_in_bytes(
            n_synapse_types, self.__timing_dependence.n_weight_terms)
        if self.__neuromodulation:
            size += self.__neuromodulation.get_parameters_sdram_usage_in_bytes(
                n_neurons, n_synapse_types)
        return size

    @overrides(AbstractPlasticSynapseDynamics.write_parameters)
    def write_parameters(
            self, spec: DataSpecificationBase, region: int,
            global_weight_scale: float,
            synapse_weight_scales: NDArray[floating]) -> None:
        spec.comment("Writing Plastic Parameters")

        # Switch focus to the region:
        spec.switch_write_focus(region)

        # Whether to use back-prop delay
        spec.write_value(int(self.__backprop_delay))

        # Write timing dependence parameters to region
        self.__timing_dependence.write_parameters(
            spec, global_weight_scale, synapse_weight_scales)

        # Write weight dependence information to region
        self.__weight_dependence.write_parameters(
            spec, global_weight_scale, synapse_weight_scales,
            self.__timing_dependence.n_weight_terms)

        if self.__neuromodulation:
            self.__neuromodulation.write_parameters(
                spec, region, global_weight_scale, synapse_weight_scales)

    @property
    def _n_header_bytes(self) -> int:
        # The header contains a single timestamp and pre-trace
        n_bytes = (
            TIME_STAMP_BYTES + self.__timing_dependence.pre_trace_n_bytes)

        # The actual number of bytes is in a word-aligned struct, so work out
        # the number of bytes as a number of words
        return int(math.ceil(float(n_bytes) / BYTES_PER_WORD)) * BYTES_PER_WORD

    def __get_n_connections(
            self, n_connections: int, check_length_padded: bool = True) -> int:
        synapse_structure = self.__timing_dependence.synaptic_structure
        if self.__pad_to_length is not None and check_length_padded:
            n_connections = max(n_connections, self.__pad_to_length)
        if n_connections == 0:
            return 0
        # 2 == two half words per word
        fp_size_words = (
            n_connections // 2 if n_connections % 2 == 0
            else (n_connections + 1) // 2)
        pp_size_bytes = (
            self._n_header_bytes +
            (synapse_structure.get_n_half_words_per_connection() *
             BYTES_PER_SHORT * n_connections))
        # Neuromodulation synapses have the actual weight separately
        if self.__neuromodulation:
            pp_size_bytes += BYTES_PER_SHORT * n_connections
        pp_size_words = int(math.ceil(float(pp_size_bytes) / BYTES_PER_WORD))

        return fp_size_words + pp_size_words

    @overrides(AbstractPlasticSynapseDynamics.
               get_n_words_for_plastic_connections)
    def get_n_words_for_plastic_connections(self, n_connections: int) -> int:
        """
        :param n_connections:
        """
        return self.__get_n_connections(n_connections)

    @overrides(AbstractPlasticSynapseDynamics.get_plastic_synaptic_data)
    def get_plastic_synaptic_data(
            self, connections: ConnectionsArray,
            connection_row_indices: NDArray[integer], n_rows: int,
            n_synapse_types: int,
            max_n_synapses: int, max_atoms_per_core: int) -> Tuple[
                List[NDArray[uint32]], List[NDArray[uint32]],
                NDArray[uint32], NDArray[uint32]]:
        n_synapse_type_bits = get_n_bits(n_synapse_types)
        n_neuron_id_bits = get_n_bits(max_atoms_per_core)
        neuron_id_mask = (1 << n_neuron_id_bits) - 1

        # Get the fixed data
        fixed_plastic = (
            (connections["delay"].astype(uint16) <<
             (n_neuron_id_bits + n_synapse_type_bits)) |
            (connections["synapse_type"].astype(uint16)
             << n_neuron_id_bits) |
            (connections["target"].astype(uint16) & neuron_id_mask))
        fixed_plastic_rows = self.convert_per_connection_data_to_rows(
            connection_row_indices, n_rows,
            fixed_plastic.view(dtype=uint8).reshape((-1, 2)),
            max_n_synapses)
        fp_size = self.get_n_items(fixed_plastic_rows, BYTES_PER_SHORT)
        if self.__pad_to_length is not None:
            # Pad the data
            fixed_plastic_rows = self._pad_row(
                fixed_plastic_rows, BYTES_PER_SHORT)
        fp_data = self.get_words(fixed_plastic_rows)

        # Get the plastic data by inserting the weight into the half-word
        # specified by the synapse structure
        synapse_structure = self.__timing_dependence.synaptic_structure
        n_half_words = synapse_structure.get_n_half_words_per_connection()
        half_word = synapse_structure.get_weight_half_word()
        # If neuromodulation, the real weight comes first
        if self.__neuromodulation:
            n_half_words += 1
            half_word = 0
        plastic_plastic = numpy.zeros(
            len(connections) * n_half_words, dtype=uint16)
        plastic_plastic[half_word::n_half_words] = \
            numpy.rint(numpy.abs(connections["weight"])).astype(uint16)

        # Convert the plastic data into groups of bytes per connection and
        # then into rows
        plastic_plastic_bytes = plastic_plastic.view(dtype=uint8).reshape(
            (-1, n_half_words * BYTES_PER_SHORT))
        plastic_plastic_row_data = self.convert_per_connection_data_to_rows(
            connection_row_indices, n_rows, plastic_plastic_bytes,
            max_n_synapses)

        # pp_size = fp_size in words => fp_size * no_bytes / 4 (bytes)
        if self.__pad_to_length is not None:
            # Pad the data
            plastic_plastic_row_data = self._pad_row(
                plastic_plastic_row_data, n_half_words * BYTES_PER_SHORT)
        plastic_headers = numpy.zeros(
            (n_rows, self._n_header_bytes), dtype=uint8)
        plastic_plastic_rows = [
            numpy.concatenate((
                plastic_headers[i], plastic_plastic_row_data[i]))
            for i in range(n_rows)]
        pp_size = self.get_n_items(plastic_plastic_rows, BYTES_PER_WORD)
        pp_data = self.get_words(plastic_plastic_rows)

        return fp_data, pp_data, fp_size, pp_size

    def _pad_row(self, rows: List[NDArray],
                 no_bytes_per_connection: int) -> List[NDArray]:
        pad_len = self.__pad_to_length or 1
        # Row elements are (individual) bytes
        return [
            numpy.concatenate((
                row, numpy.zeros(
                    numpy.clip(
                        no_bytes_per_connection * pad_len - row.size,
                        0, None)).astype(dtype=uint8))
                ).view(dtype=uint8)
            for row in rows]

    @overrides(
        AbstractPlasticSynapseDynamics.get_n_plastic_plastic_words_per_row)
    def get_n_plastic_plastic_words_per_row(
            self, pp_size: NDArray[uint32]) -> NDArray[integer]:
        # pp_size is in words, so return
        return pp_size

    @overrides(
        AbstractPlasticSynapseDynamics.get_n_fixed_plastic_words_per_row)
    def get_n_fixed_plastic_words_per_row(
            self, fp_size: NDArray[uint32]) -> NDArray[integer]:
        # fp_size is in half-words
        return numpy.ceil(fp_size / 2.0).astype(dtype=uint32)

    @overrides(AbstractPlasticSynapseDynamics.get_n_synapses_in_rows)
    def get_n_synapses_in_rows(self, pp_size: NDArray[uint32],
                               fp_size: NDArray[uint32]) -> NDArray[integer]:
        # Each fixed-plastic synapse is a half-word and fp_size is in half
        # words so just return it
        return fp_size

    @overrides(AbstractPlasticSynapseDynamics.read_plastic_synaptic_data)
    def read_plastic_synaptic_data(
            self, n_synapse_types: int, pp_size: NDArray[uint32],
            pp_data: List[NDArray[uint32]], fp_size: NDArray[uint32],
            fp_data: List[NDArray[uint32]],
            max_atoms_per_core: int) -> ConnectionsArray:
        n_rows = len(fp_size)

        n_synapse_type_bits = get_n_bits(n_synapse_types)
        n_neuron_id_bits = get_n_bits(max_atoms_per_core)
        neuron_id_mask = (1 << n_neuron_id_bits) - 1

        data_fixed = numpy.concatenate([
            fp_data[i].view(dtype=uint16)[0:fp_size[i]]
            for i in range(n_rows)])
        pp_without_headers = [
            row.view(dtype=uint8)[self._n_header_bytes:] for row in pp_data]
        synapse_structure = self.__timing_dependence.synaptic_structure
        n_half_words = synapse_structure.get_n_half_words_per_connection()
        half_word = synapse_structure.get_weight_half_word()
        if self.__neuromodulation:
            n_half_words += 1
            half_word = 0
        pp_half_words = numpy.concatenate([
            pp[:size * n_half_words * BYTES_PER_SHORT].view(uint16)[
                half_word::n_half_words]
            for pp, size in zip(pp_without_headers, fp_size)])

        connections = numpy.zeros(
            data_fixed.size, dtype=NUMPY_CONNECTORS_DTYPE)
        connections["source"] = numpy.concatenate(
            [numpy.repeat(i, fp_size[i]) for i in range(len(fp_size))])
        connections["target"] = data_fixed & neuron_id_mask
        connections["weight"] = pp_half_words
        connections["delay"] = data_fixed >> (
            n_neuron_id_bits + n_synapse_type_bits)
        return connections

    @overrides(AbstractPlasticSynapseDynamics.get_weight_mean)
    def get_weight_mean(self, connector: AbstractConnector,
                        synapse_info: SynapseInformation) -> float:
        # Because the weights could all be changed to the maximum, the mean
        # has to be given as the maximum for scaling
        return self.get_weight_maximum(connector, synapse_info)

    @overrides(AbstractPlasticSynapseDynamics.get_weight_variance)
    def get_weight_variance(
           self, connector: AbstractConnector, weights: WEIGHTS,
            synapse_info: SynapseInformation) -> float:
        # Because the weights could all be changed to the maximum, the variance
        # has to be given as no variance
        return 0.0

    @overrides(AbstractPlasticSynapseDynamics.get_weight_maximum)
    def get_weight_maximum(self, connector: AbstractConnector,
                           synapse_info: SynapseInformation) -> float:
        w_max = super().get_weight_maximum(connector, synapse_info)
        # The maximum weight is the largest that it could be set to from
        # the weight dependence
        return max(w_max, self.__weight_dependence.weight_maximum)

    @overrides(AbstractPlasticSynapseDynamics.get_parameter_names)
    def get_parameter_names(self) -> Iterable[str]:
        yield 'weight'
        yield 'delay'
        yield from self.__timing_dependence.get_parameter_names()
        yield from self.__weight_dependence.get_parameter_names()

    @overrides(AbstractPlasticSynapseDynamics.get_max_synapses)
    def get_max_synapses(self, n_words: int) -> int:
        # Subtract the header size that will always exist
        n_header_words = self._n_header_bytes // BYTES_PER_WORD
        n_words_space = n_words - n_header_words

        # Get plastic plastic size per connection
        synapse_structure = self.__timing_dependence.synaptic_structure
        bytes_per_pp = (
            synapse_structure.get_n_half_words_per_connection() *
            BYTES_PER_SHORT)
        if self.__neuromodulation:
            bytes_per_pp += BYTES_PER_SHORT

        # The fixed plastic size per connection is 2 bytes
        bytes_per_fp = BYTES_PER_SHORT

        # Maximum possible connections, ignoring word alignment
        n_connections = (n_words_space * BYTES_PER_WORD) // (
            bytes_per_pp + bytes_per_fp)

        check_length_padded = False

        # Reduce until correct
        while (self.__get_n_connections(n_connections, check_length_padded) >
               n_words):
            n_connections -= 1

        return n_connections

    @property
    @overrides(AbstractGenerateOnMachine.gen_matrix_id)
    def gen_matrix_id(self) -> int:
        return MatrixGeneratorID.STDP_MATRIX.value

    @overrides(AbstractGenerateOnMachine.gen_matrix_params)
    def gen_matrix_params(
            self, synaptic_matrix_offset: int, delayed_matrix_offset: int,
            app_edge: ProjectionApplicationEdge,
            synapse_info: SynapseInformation, max_row_info: MaxRowInfo,
            max_pre_atoms_per_core: int, max_post_atoms_per_core: int
            ) -> NDArray[uint32]:
        vertex = app_edge.post_vertex
        n_synapse_type_bits = get_n_bits(
            vertex.neuron_impl.get_n_synapse_types())
        n_synapse_index_bits = get_n_bits(max_post_atoms_per_core)
        max_delay = app_edge.post_vertex.splitter.max_support_delay()
        max_delay_bits = get_n_bits(max_delay)
        synapse_struct = self.__timing_dependence.synaptic_structure
        n_half_words = synapse_struct.get_n_half_words_per_connection()
        half_word = synapse_struct.get_weight_half_word()
        if self.__neuromodulation:
            n_half_words += 1
            half_word = 0
        write_row_number_to_header = 0
        row_offset = 0
        return numpy.array([
            synaptic_matrix_offset, delayed_matrix_offset,
            max_row_info.undelayed_max_n_synapses,
            max_row_info.delayed_max_n_synapses,
            max_row_info.undelayed_max_words, max_row_info.delayed_max_words,
            synapse_info.synapse_type, n_synapse_type_bits,
            n_synapse_index_bits, app_edge.n_delay_stages + 1,
            max_delay, max_delay_bits, app_edge.pre_vertex.n_atoms,
            max_pre_atoms_per_core, self._n_header_bytes // BYTES_PER_SHORT,
            n_half_words, half_word, write_row_number_to_header, row_offset],
            dtype=uint32)

    @property
    @overrides(AbstractGenerateOnMachine.gen_matrix_params_size_in_bytes)
    def gen_matrix_params_size_in_bytes(self) -> int:
        return 19 * BYTES_PER_WORD

    @property
    @overrides(AbstractPlasticSynapseDynamics.changes_during_run)
    def changes_during_run(self) -> bool:
        return True

    @property
    @overrides(AbstractPlasticSynapseDynamics.is_combined_core_capable)
    def is_combined_core_capable(self) -> bool:
        return self.__neuromodulation is None

    @property
    @overrides(AbstractPlasticSynapseDynamics.is_split_core_capable)
    def is_split_core_capable(self) -> bool:
        return True

    @property
    @overrides(AbstractPlasticSynapseDynamics.pad_to_length)
    def pad_to_length(self) -> Optional[int]:
        return self.__pad_to_length

    @property
    @overrides(AbstractPlasticSynapseDynamics.synapses_per_second)
    def synapses_per_second(self) -> int:
        # From Synapse-Centric Mapping of Cortical Models to the SpiNNaker
        # Neuromorphic Architecture
        return 1400000



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable
from spinn_utilities.overrides import overrides
from spinn_front_end_common.interface.ds import DataSpecificationBase
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD
from .abstract_elimination import AbstractElimination


class RandomByWeightElimination(AbstractElimination):
    """
    Elimination Rule that depends on the weight of a synapse.
    """

    __slots__ = (
        "__prob_elim_depressed",
        "__prob_elim_potentiated",
        "__threshold")

    def __init__(
            self, threshold: float, prob_elim_depressed: float = 0.0245,
            prob_elim_potentiated: float = 1.36e-4):
        """
        :param threshold:
            Below this weight is considered depression, above or equal to this
            weight is considered potentiation (or the static weight of the
            connection on static weight connections)
        :param prob_elim_depressed:
            The probability of elimination if the weight has been depressed
            (ignored on static weight connections)
        :param prob_elim_potentiated:
            The probability of elimination of the weight has been potentiated
            or has not changed (and also used on static weight connections)
        """
        self.__prob_elim_depressed = prob_elim_depressed
        self.__prob_elim_potentiated = prob_elim_potentiated
        self.__threshold = threshold

    @property
    @overrides(AbstractElimination.vertex_executable_suffix)
    def vertex_executable_suffix(self) -> str:
        return "_weight"

    @overrides(AbstractElimination.get_parameters_sdram_usage_in_bytes)
    def get_parameters_sdram_usage_in_bytes(self) -> int:
        return 3 * BYTES_PER_WORD

    @overrides(AbstractElimination.write_parameters)
    def write_parameters(
            self, spec: DataSpecificationBase, weight_scale: float) -> None:
        spec.write_value(int(self.__prob_elim_depressed * 0xFFFFFFFF))
        spec.write_value(int(self.__prob_elim_potentiated * 0xFFFFFFFF))
        spec.write_value(self.__threshold * weight_scale)

    @overrides(AbstractElimination.get_parameter_names)
    def get_parameter_names(self) -> Iterable[str]:
        return ("prob_elim_depressed", "prob_elim_potentiated", "threshold")



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from spynnaker.pyNN.protocols import RetinaKey


class PushBotRetinaResolution(Enum):
    """
    Resolutions supported by the pushbot retina device.
    """

    #: The native resolution
    NATIVE_128_X_128 = RetinaKey.NATIVE_128_X_128
    #: Down sampled 4 (:math:`2 \times 2`) pixels to 1
    DOWNSAMPLE_64_X_64 = RetinaKey.DOWNSAMPLE_64_X_64
    #: Down sampled 16 (:math:`4 \times 4`) pixels to 1
    DOWNSAMPLE_32_X_32 = RetinaKey.DOWNSAMPLE_32_X_32
    #: Down sampled 64 (:math:`8 \times 8`) pixels to 1
    DOWNSAMPLE_16_X_16 = RetinaKey.DOWNSAMPLE_16_X_16



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spynnaker.pyNN.models.neuron.implementations import (
    AbstractStandardNeuronComponent)


class AbstractInputType(
        AbstractStandardNeuronComponent, metaclass=AbstractBase):
    """
    Represents a possible input type for a neuron model (e.g., current).
    """
    __slots__ = ()

    @abstractmethod
    def get_global_weight_scale(self) -> float:
        """
        Get the global weight scaling value.

        :return: The global weight scaling value
        """
        raise NotImplementedError



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from math import ceil
from typing import (
    Dict, Iterable, List, Tuple, cast, TYPE_CHECKING)

import numpy
from numpy import floating, uint32
from numpy.typing import NDArray

from spinn_utilities.overrides import overrides

from pacman.model.graphs.application import ApplicationVertex

from spinn_front_end_common.interface.ds import (
    DataType, DataSpecificationGenerator)
from spinn_front_end_common.utilities.constants import (
    BYTES_PER_SHORT, BYTES_PER_WORD)

from spynnaker.pyNN.exceptions import SynapticConfigurationException
from spynnaker.pyNN.models.abstract_models import ColouredApplicationVertex
from spynnaker.pyNN.models.neural_projections.connectors import (
    ConvolutionConnector, AbstractConnector)
from spynnaker.pyNN.models.neuron.synapse_dynamics import (
    AbstractSupportsSignedWeights)
from spynnaker.pyNN.types import WEIGHTS_DELAYS_IN
from spynnaker.pyNN.models.common.local_only_2d_common import (
    get_div_const, get_rinfo_for_spike_source, get_sources_for_target,
    BITS_PER_SHORT, N_COLOUR_BITS_BITS, KEY_INFO_SIZE,
    get_first_and_last_slice, Source)
from .abstract_local_only import AbstractLocalOnly

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neuron.population_vertex import (
        PopulationVertex)
    from spynnaker.pyNN.models.projection import Projection
    from spynnaker.pyNN.models.neuron import (
        PopulationMachineLocalOnlyCombinedVertex)
    from spynnaker.pyNN.models.neuron.synapse_dynamics import (
        AbstractSynapseDynamics)


#: Size of convolution config main bytes
CONV_CONFIG_SIZE = (6 * BYTES_PER_SHORT) + (4 * BYTES_PER_WORD)

#: Size of source information
SOURCE_INFO_SIZE = KEY_INFO_SIZE + (6 * BYTES_PER_WORD)


class LocalOnlyConvolution(AbstractLocalOnly, AbstractSupportsSignedWeights):
    """
    A convolution synapse dynamics that can process spikes with only DTCM.
    """

    __slots__ = (
        "__cached_sources",
    )

    def __init__(self, delay: WEIGHTS_DELAYS_IN = None):

        """
        :param delay:
            The delay used in the connection; by default 1 time step
        """
        super().__init__(delay)

        # Store the sources to avoid recalculation
        self.__cached_sources: Dict[ApplicationVertex, Dict[
                Tuple[ColouredApplicationVertex, str],
                List[Source]]] = dict()

    @property
    def _delay(self) -> float:
        # Guaranteed by check in init
        return cast(float, self.delay)

    @overrides(AbstractLocalOnly.merge)
    def merge(self, synapse_dynamics: AbstractSynapseDynamics
              ) -> LocalOnlyConvolution:
        if not isinstance(synapse_dynamics, LocalOnlyConvolution):
            raise SynapticConfigurationException(
                "All targets of this Population must have a synapse_type of"
                " Convolution")
        return synapse_dynamics

    @overrides(AbstractLocalOnly.get_vertex_executable_suffix)
    def get_vertex_executable_suffix(self) -> str:
        return "_conv"

    @property
    @overrides(AbstractLocalOnly.changes_during_run)
    def changes_during_run(self) -> bool:
        return False

    @overrides(AbstractLocalOnly.get_parameters_usage_in_bytes)
    def get_parameters_usage_in_bytes(
            self, n_atoms: int,
            incoming_projections: Iterable[Projection]) -> int:
        # pylint: disable=protected-access
        n_bytes = 0
        kernel_bytes = 0
        connectors_seen = set()
        edges_seen = set()
        for incoming in incoming_projections:
            s_info = incoming._synapse_information
            if not isinstance(s_info.connector, ConvolutionConnector):
                raise SynapticConfigurationException(
                    "Only ConvolutionConnector can be used with a synapse type"
                    " of Convolution")
            app_edge = incoming._projection_edge
            if app_edge not in edges_seen:
                edges_seen.add(app_edge)
                n_bytes += SOURCE_INFO_SIZE
            if s_info.connector not in connectors_seen:
                connectors_seen.add(s_info.connector)
                kernel_bytes += s_info.connector.kernel_n_bytes
            n_bytes += s_info.connector.parameters_n_bytes

        if kernel_bytes % BYTES_PER_WORD != 0:
            kernel_bytes += BYTES_PER_SHORT

        return CONV_CONFIG_SIZE + n_bytes + kernel_bytes

    @overrides(AbstractLocalOnly.write_parameters)
    def write_parameters(
            self, spec: DataSpecificationGenerator, region: int,
            machine_vertex: PopulationMachineLocalOnlyCombinedVertex,
            weight_scales: NDArray[floating]) -> None:
        # pylint: disable=protected-access

        # Get incoming sources for this vertex
        app_vertex = machine_vertex._pop_vertex
        sources = self.__get_sources_for_target(app_vertex)

        size = self.get_parameters_usage_in_bytes(
            machine_vertex.vertex_slice.n_atoms,
            app_vertex.incoming_projections)
        spec.reserve_memory_region(region, size, label="LocalOnlyConvolution")
        spec.switch_write_focus(region)

        # Get spec for each incoming source
        connector_weight_index: Dict[AbstractConnector, int] = dict()
        next_weight_index: int = 0
        source_data = list()
        connector_data: List[NDArray[uint32]] = list()
        weight_data = list()
        for (pre_vertex, part_id), source_infos in sources.items():

            # Add connectors as needed
            first_conn_index = len(connector_data)
            for source in source_infos:
                # pylint: disable=protected-access
                conn = cast(
                    ConvolutionConnector,
                    source.projection._synapse_information.connector)
                app_edge = source.projection._projection_edge

                # Work out whether the connector needs a new weight index
                if conn in connector_weight_index:
                    weight_index = connector_weight_index[conn]
                else:
                    weight_index = next_weight_index
                    connector_weight_index[conn] = weight_index
                    next_weight_index += conn.kernel_n_weights
                    weight_data.append(conn.get_encoded_kernel_weights(
                        app_edge, weight_scales))

                connector_data.append(conn.get_local_only_data(
                    app_edge, source.local_delay, source.delay_stage,
                    weight_index))

            # Get the source routing information
            r_info, core_mask, mask_shift = get_rinfo_for_spike_source(
                pre_vertex, part_id)

            # Get the width / height per core / last_core
            first_slice, last_slice = get_first_and_last_slice(pre_vertex)
            width_per_core = first_slice.shape[0]
            height_per_core = first_slice.shape[1]
            width_on_last_core = last_slice.shape[0]
            height_on_last_core = last_slice.shape[1]

            # Get cores per width / height
            pre_shape = list(pre_vertex.atoms_shape)
            cores_per_width = int(ceil(pre_shape[0] / width_per_core))
            cores_per_height = int(ceil(pre_shape[1] / height_per_core))

            # Add the key and mask...
            source_data.extend([r_info.key, r_info.mask])
            # ... start connector index, n_colour_bits, count of connectors ...
            source_data.append(
                (len(source_infos) << BITS_PER_SHORT) +
                (pre_vertex.n_colour_bits <<
                 (BITS_PER_SHORT - N_COLOUR_BITS_BITS)) +
                first_conn_index)
            # ... core mask, mask shift ...
            source_data.append((mask_shift << BITS_PER_SHORT) + core_mask)
            # ... height / width per core ...
            source_data.append(
                (width_per_core << BITS_PER_SHORT) + height_per_core)
            # ... height / width last core ...
            source_data.append(
                (width_on_last_core << BITS_PER_SHORT) + height_on_last_core)
            # ... cores per height / width ...
            source_data.append(
                (cores_per_width << BITS_PER_SHORT) + cores_per_height)
            # ... 1 / width per core ...
            source_data.append(get_div_const(width_per_core))
            # ... 1 / width last core ...
            source_data.append(get_div_const(width_on_last_core))
            # ... 1 / cores_per_width
            source_data.append(get_div_const(cores_per_width))

        if next_weight_index % 2 != 0:
            weight_data.append(numpy.array([0], dtype="int16"))

        # Write the common spec
        post_slice = machine_vertex.vertex_slice
        post_start = numpy.array(post_slice.start)
        post_shape = numpy.array(post_slice.shape)
        post_end = (post_start + post_shape) - 1
        spec.write_value(post_start[1], data_type=DataType.INT16)
        spec.write_value(post_start[0], data_type=DataType.INT16)
        spec.write_value(post_end[1], data_type=DataType.INT16)
        spec.write_value(post_end[0], data_type=DataType.INT16)
        spec.write_value(post_shape[1], data_type=DataType.INT16)
        spec.write_value(post_shape[0], data_type=DataType.INT16)
        spec.write_value(len(sources))
        spec.write_value(len(connector_data))
        spec.write_value(next_weight_index)

        # Write the data
        # pylint: disable=unexpected-keyword-arg
        spec.write_array(numpy.array(source_data, dtype="uint32"))
        spec.write_array(numpy.concatenate(connector_data, dtype="uint32"))
        spec.write_array(
            numpy.concatenate(weight_data, dtype="int16").view("uint32"))

    def __get_sources_for_target(
            self, app_vertex: PopulationVertex) -> Dict[
                Tuple[ColouredApplicationVertex, str], List[Source]]:
        """
        Get all the application vertex sources that will hit the given
        application vertex.

        :param app_vertex: The vertex being targeted
        :return:
            A dict of source PopulationApplicationVertex to list of source
            information
        """
        sources = self.__cached_sources.get(app_vertex)
        if sources is None:
            sources = get_sources_for_target(app_vertex)
            self.__cached_sources[app_vertex] = sources
        return sources

    @staticmethod
    def __connector(projection: Projection) -> ConvolutionConnector:
        # pylint: disable=protected-access
        return cast(ConvolutionConnector,
                    projection._synapse_information.connector)

    @staticmethod
    def __get_synapse_type(proj: Projection, target: str) -> int:
        edge = proj._projection_edge  # pylint: disable=protected-access
        synapse_type = edge.post_vertex.get_synapse_id_by_target(target)
        # Checked during connection validation, assumed constant
        assert synapse_type is not None
        return synapse_type

    @overrides(AbstractSupportsSignedWeights.get_positive_synapse_index)
    def get_positive_synapse_index(
            self, incoming_projection: Projection) -> int:
        return self.__get_synapse_type(
            incoming_projection,
            self.__connector(incoming_projection).positive_receptor_type)

    @overrides(AbstractSupportsSignedWeights.get_negative_synapse_index)
    def get_negative_synapse_index(
            self, incoming_projection: Projection) -> int:
        return self.__get_synapse_type(
            incoming_projection,
            self.__connector(incoming_projection).negative_receptor_type)

    @overrides(AbstractSupportsSignedWeights.get_maximum_positive_weight)
    def get_maximum_positive_weight(
            self, incoming_projection: Projection) -> float:
        conn = self.__connector(incoming_projection)
        # We know the connector doesn't care about the argument
        # conn.kernel_weights known to be an array of floats
        max_weight = cast(float, numpy.amax(conn.kernel_weights))
        return max_weight if max_weight > 0 else 0

    @overrides(AbstractSupportsSignedWeights.get_minimum_negative_weight)
    def get_minimum_negative_weight(
            self, incoming_projection: Projection) -> float:
        conn = self.__connector(incoming_projection)
        # This is different because the connector happens to support this
        # conn.kernel_weights known to be an array of floats
        min_weight = cast(float, numpy.amin(conn.kernel_weights))
        return min_weight if min_weight < 0 else 0

    @overrides(AbstractSupportsSignedWeights.get_mean_positive_weight)
    def get_mean_positive_weight(
            self, incoming_projection: Projection) -> float:
        conn = self.__connector(incoming_projection)
        pos_weights = conn.kernel_weights[conn.kernel_weights > 0]
        if len(pos_weights) == 0:
            return 0
        return float(numpy.mean(pos_weights))

    @overrides(AbstractSupportsSignedWeights.get_mean_negative_weight)
    def get_mean_negative_weight(
            self, incoming_projection: Projection) -> float:
        conn = self.__connector(incoming_projection)
        neg_weights = conn.kernel_weights[conn.kernel_weights < 0]
        if len(neg_weights) == 0:
            return 0
        return float(numpy.mean(neg_weights))

    @overrides(AbstractSupportsSignedWeights.get_variance_positive_weight)
    def get_variance_positive_weight(
            self, incoming_projection: Projection) -> float:
        conn = self.__connector(incoming_projection)
        pos_weights = conn.kernel_weights[conn.kernel_weights > 0]
        if len(pos_weights) == 0:
            return 0
        return float(numpy.var(pos_weights))

    @overrides(AbstractSupportsSignedWeights.get_variance_negative_weight)
    def get_variance_negative_weight(
            self, incoming_projection: Projection) -> float:
        conn = self.__connector(incoming_projection)
        neg_weights = conn.kernel_weights[conn.kernel_weights < 0]
        if len(neg_weights) == 0:
            return 0
        return float(numpy.var(neg_weights))

    @property
    @overrides(AbstractLocalOnly.synapses_per_second)
    def synapses_per_second(self) -> int:
        # This is a guess!
        return 15000000



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .push_bot_laser_device import PushBotSpiNNakerLinkLaserDevice
from .push_bot_led_device import PushBotSpiNNakerLinkLEDDevice
from .push_bot_motor_device import PushBotSpiNNakerLinkMotorDevice
from .push_bot_retina_device import PushBotSpiNNakerLinkRetinaDevice
from .push_bot_speaker_device import PushBotSpiNNakerLinkSpeakerDevice

__all__ = ["PushBotSpiNNakerLinkLaserDevice",
           "PushBotSpiNNakerLinkLEDDevice",
           "PushBotSpiNNakerLinkMotorDevice",
           "PushBotSpiNNakerLinkRetinaDevice",
           "PushBotSpiNNakerLinkSpeakerDevice"]



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import (
    Iterable, List, Optional, Sequence, Tuple, TYPE_CHECKING, Union)
from spinn_utilities.overrides import overrides
from pacman.model.graphs.application import (
    ApplicationVertex, ApplicationVirtualVertex)
from pacman.model.routing_info import BaseKeyAndMask
from spinn_front_end_common.abstract_models import (
    AbstractVertexWithEdgeToDependentVertices, HasCustomAtomKeyMap)
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spynnaker.pyNN.models.neuron import PopulationVertex
from .abstract_ethernet_controller import AbstractEthernetController
from .abstract_multicast_controllable_device import (
    AbstractMulticastControllableDevice)
if TYPE_CHECKING:
    from pacman.model.graphs.machine.machine_vertex import MachineVertex
    from pacman.model.routing_info.routing_info import RoutingInfo
    from spynnaker.pyNN.models.neuron.implementations import AbstractNeuronImpl
    from spynnaker.pyNN.models.neuron import AbstractPyNNNeuronModel
    from .abstract_ethernet_translator import AbstractEthernetTranslator
    from spynnaker.pyNN.extra_algorithms.splitter_components import (
        SplitterPopulationVertex)


class ExternalDeviceLifControlVertex(
        PopulationVertex,
        AbstractEthernetController,
        AbstractVertexWithEdgeToDependentVertices,
        HasCustomAtomKeyMap):
    """
    Abstract control module for the pushbot, based on the LIF neuron, but
    without spikes, and using the voltage as the output to the various devices.
    """
    __slots__ = (
        "__dependent_vertices",
        "__devices",
        "__indices",
        "__message_translator")

    # all commands will use this mask
    _DEFAULT_COMMAND_MASK = 0xFFFFFFFF

    def __init__(
            self, *, devices: Sequence[AbstractMulticastControllableDevice],
            create_edges: bool,
            max_atoms_per_core: Union[int, Tuple[int, ...]],
            n_synapse_cores: Optional[int],
            allow_delay_extensions: bool,
            neuron_impl: AbstractNeuronImpl,
            pynn_model: AbstractPyNNNeuronModel,
            translator: Optional[AbstractEthernetTranslator] = None,
            spikes_per_second: Optional[float] = None,
            label: Optional[str] = None,
            ring_buffer_sigma: Optional[float] = None,
            max_expected_summed_weight: Optional[List[float]] = None,
            incoming_spike_buffer_size: Optional[int] = None,
            drop_late_spikes: Optional[bool] = None,
            splitter: Optional[SplitterPopulationVertex] = None,
            seed: Optional[int] = None, n_colour_bits: Optional[int] = None):
        """
        :param devices:
            The AbstractMulticastControllableDevice instances to be controlled
            by the population
        :param create_edges:
            True if edges to the devices should be added by this device (set
            to False if using the device over Ethernet using a translator)
        :param max_atoms_per_core:
        :param neuron_impl:
        :param pynn_model:
        :param translator:
            Translator to be used when used for Ethernet communication.  Must
            be provided if the device is to be controlled over Ethernet.
        :param spikes_per_second:
        :param label:
        :param ring_buffer_sigma:
        :param incoming_spike_buffer_size:
        :param splitter: splitter from application vertices to machine vertices
        :param n_colour_bits: The number of colour bits to use
        """
        if drop_late_spikes is None:
            drop_late_spikes = False
        extra_partition_ids = [
            dev.device_control_partition_id for dev in devices]
        super().__init__(
            n_neurons=len(devices),
            label=f"ext_dev{devices}" if label is None else label,
            max_atoms_per_core=max_atoms_per_core,
            n_synapse_cores=n_synapse_cores,
            allow_delay_extensions=allow_delay_extensions,
            spikes_per_second=spikes_per_second,
            ring_buffer_sigma=ring_buffer_sigma,
            max_expected_summed_weight=max_expected_summed_weight,
            incoming_spike_buffer_size=incoming_spike_buffer_size,
            neuron_impl=neuron_impl, pynn_model=pynn_model,
            drop_late_spikes=drop_late_spikes, splitter=splitter, seed=seed,
            n_colour_bits=n_colour_bits, extra_partitions=extra_partition_ids)

        if not devices:
            raise ConfigurationException("No devices specified")

        self.__devices = {dev.device_control_partition_id: dev
                          for dev in devices}
        self.__indices = {dev.device_control_partition_id: i
                          for i, dev in enumerate(devices)}
        self.__message_translator = translator

        # Add the edges to the devices if required
        self.__dependent_vertices = (
            self.__dependents(devices) if create_edges else ())

    @staticmethod
    def __dependents(
            devices: Sequence[AbstractMulticastControllableDevice]) -> Tuple[
                ApplicationVirtualVertex, ...]:
        return tuple(
            dev for dev in devices
            if isinstance(dev, ApplicationVirtualVertex))

    @overrides(AbstractVertexWithEdgeToDependentVertices.dependent_vertices)
    def dependent_vertices(self) -> Iterable[ApplicationVertex]:
        return self.__dependent_vertices

    @overrides(AbstractVertexWithEdgeToDependentVertices
               .edge_partition_identifiers_for_dependent_vertex)
    def edge_partition_identifiers_for_dependent_vertex(
            self, vertex: ApplicationVertex) -> Iterable[str]:
        assert isinstance(vertex, AbstractMulticastControllableDevice)
        return [vertex.device_control_partition_id]

    @overrides(AbstractEthernetController.get_external_devices)
    def get_external_devices(self) -> Iterable[
            AbstractMulticastControllableDevice]:
        return self.__devices.values()

    @overrides(AbstractEthernetController.get_message_translator)
    def get_message_translator(self) -> AbstractEthernetTranslator:
        if self.__message_translator is None:
            raise ConfigurationException(
                "This population was not given a translator, and so cannot be"
                "used for Ethernet communication.  Please provide a "
                "translator for the population.")
        return self.__message_translator

    @overrides(AbstractEthernetController.get_outgoing_partition_ids)
    def get_outgoing_partition_ids(self) -> List[str]:
        return list(self.__devices.keys())

    @overrides(HasCustomAtomKeyMap.get_atom_key_map)
    def get_atom_key_map(
            self, pre_vertex: MachineVertex, partition_id: str,
            routing_info: RoutingInfo) -> Iterable[Tuple[int, int]]:
        index = self.__indices[partition_id]
        device = self.__devices[partition_id]
        return [(index, device.device_control_key)]

    @overrides(PopulationVertex.get_fixed_key_and_mask)
    def get_fixed_key_and_mask(
            self, partition_id: str) -> Optional[BaseKeyAndMask]:
        return BaseKeyAndMask(
            self.__devices[partition_id].device_control_key,
            self._DEFAULT_COMMAND_MASK)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional
import scipy.stats
from spinn_utilities.overrides import overrides
from pyNN.random import RandomDistribution
from spynnaker.pyNN.utilities.random_stats import AbstractRandomStats


class RandomStatsScipyImpl(AbstractRandomStats):
    """
    A Random Statistics object that uses scipy directly.
    """

    def __init__(self, distribution_type: str):
        """
        :param distribution_type: Name of the distribution type.
        """
        self._scipy_stats = getattr(scipy.stats, distribution_type)

    @overrides(AbstractRandomStats.cdf)
    def cdf(self, dist: RandomDistribution, v: float) -> float:
        return self._scipy_stats.cdf(v, *dist.parameters)

    @overrides(AbstractRandomStats.ppf)
    def ppf(self, dist: RandomDistribution, p: float) -> float:
        return self._scipy_stats.ppf(p, *dist.parameters)

    @overrides(AbstractRandomStats.mean)
    def mean(self, dist: RandomDistribution) -> float:
        return self._scipy_stats.mean(*dist.parameters)

    @overrides(AbstractRandomStats.std)
    def std(self, dist: RandomDistribution) -> float:
        return self._scipy_stats.std(*dist.parameters)

    @overrides(AbstractRandomStats.var)
    def var(self, dist: RandomDistribution) -> float:
        return self._scipy_stats.var(*dist.parameters)

    @overrides(AbstractRandomStats.high)
    def high(self, distribution: RandomDistribution) -> Optional[float]:
        if "high" in distribution.parameters:
            return distribution.parameters['high']
        return None

    @overrides(AbstractRandomStats.low)
    def low(self, distribution: RandomDistribution) -> Optional[float]:
        if "low" in distribution.parameters:
            return distribution.parameters['low']
        return None



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from pyNN.space import BaseStructure


class SupportsStructure(object, metaclass=AbstractBase):
    """
    Indicates an object that supports the setting of a PyNN structure.
    """

    @abstractmethod
    def set_structure(self, structure: BaseStructure) -> None:
        """
        Set the structure of the object.

        :param structure: The structure to set
        """
        raise NotImplementedError
